# coding=utf-8
"""
"""
from __future__ import absolute_import

from operator import attrgetter

from .base import ColumnSet


class RelatedColumnSet(ColumnSet):
    """ ColumnSet for a related entity. """

    def __init__(self, related_attr, attrs, label=None, required=False):
        """
        :param related_attr: attribute name on main entity that connects to related
        one.

        :param attrs: iterable of :class:`Column` or :class:`ColumnSet` or
                      `tuple(attribute, label, types map, col_attr)`.
        """
        self.related_attr = related_attr
        self.get_related = attrgetter(self.related_attr)

        if label is None:
            label = unicode(related_attr) \
                .replace(u'_', u' ') \
                .replace(u'.', u' ')

        self.label = self.related_label = label
        self.required = required
        ColumnSet.__init__(self, *attrs)

    def __repr__(self):
        return (u'{module}.{cls}(related_attr={attr}, label={label}, '
                u'required={required:}) at 0x{id:x}'
                u''.format(module=self.__class__.__module__,
                           cls=self.__class__.__name__,
                           attr=repr(self.related_attr),
                           label=repr(self.related_label),
                           required=repr(self.required),
                           id=id(self))).encode('utf-8')

    @property
    def attrs(self):
        for attr in ColumnSet.attrs.fget(self):
            yield u'{}.{}'.format(self.related_attr, attr)

    @property
    def labels(self):
        for label in ColumnSet.labels.fget(self):
            if not self.related_label:
                yield label
            else:
                yield u'{}:\n {}'.format(self.related_label, label)

    def data(self, item):
        # if item is None we must nonetheless call data() for all columns and
        # sub-relatedset columns, so that output values are consistent with columns
        # labels
        related = None
        if item is not None:
            try:
                related = self.get_related(item)
            except AttributeError:
                pass

        return ColumnSet.data(self, related)

    def data_for_import(self, item):
        related = None
        try:
            related = self.get_related(item)
        except AttributeError:
            pass

        return related, related


class ManyRelatedColumnSet(ColumnSet):

    def __init__(self,
                 related_attr,
                 attrs=None,
                 label=None,
                 model_cls=None,
                 form_cls=None,
                 export_label=None,
                 id_by_name_col=None,
                 manager_cls=None):
        """
        :param related_attr: attribute name on main entity that connects to related
        ones

        :param attrs: iterable of :class:`Column` or :class:`ColumnSet` or tuple
                      (attribute, label, types map).
        """
        from abilian.crm.excel import ExcelManager

        self.ID_BY_NAME_COL = id_by_name_col
        self.related_attr = related_attr
        if label is None:
            label = related_attr.replace(u'_', u' ')
        self.related_label = label
        self.model_cls = model_cls
        self.form_cls = form_cls
        self.manager_cls = manager_cls if manager_cls is not None else ExcelManager

        if export_label is None:
            export_label = label
        self.export_label = export_label

        if attrs is None:
            manager = self.create_manager()
            attrs = manager.columns.columns

        ColumnSet.__init__(self, *attrs)

    def create_manager(self):
        return self.manager_cls(self.model_cls, self.form_cls, ())

    def iter_items(self, obj):
        path = self.related_attr.split('.')

        def iter_obj(obj, attrs):
            if not attrs:
                yield obj
                raise StopIteration

            objs = getattr(obj, attrs[0])
            if not isinstance(objs, (set, list, tuple)):
                objs = (objs,)

            attrs = attrs[1:]
            for item in objs:
                # python 2.7 doesn't have "yield from"...
                for yielded_item in iter_obj(item, attrs):
                    yield yielded_item

        for item in iter_obj(obj, path):
            yield item

    @property
    def labels(self):
        for label in ColumnSet.labels.fget(self):
            if not self.related_label:
                yield label
            else:
                yield u'{}:\n {}'.format(self.related_label, label)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import cStringIO as StringIO
import logging
from itertools import ifilter
from time import gmtime, strftime

import celery
from flask import current_app, flash, render_template, request
from flask_login import current_user

from abilian.core.util import fqcn
from abilian.i18n import _, _l
from abilian.web import csrf, url_for, views
from abilian.web.action import Endpoint, FAIcon
from abilian.web.blueprints import Blueprint
from abilian.web.frontend import (ModuleAction, ModuleActionDropDown,
                                  ModuleActionGroupItem, ModuleComponent,
                                  ModuleView)
from abilian.web.util import capture_stream_errors

from .manager import ExcelManager
from .tasks import export as export_task
from .util import XLSX_MIME

logger = logging.getLogger(__name__)

bp = Blueprint('crm_excel', __name__, url_prefix='/excel')


class _ItemUpdate(object):
    """Holds item update data.

    Used in import views.

    :param item_id: primary key
    :param attrs: list of attributes, in received order
    :param signature: verify @attrs authenticity
    :param data: dict attr => new value. Keys must be in attrs
    """

    def __init__(self, item_id, attrs, signature, data):
        self.id = item_id
        self.attrs = sorted(attrs)
        self.sig = signature
        self.data = data


class BaseExcelView(ModuleView, views.View):
    excel_manager = None
    Form = None

    def __init__(self,
                 view_endpoint,
                 component='excel',
                 Form=None,
                 excel_manager=None,
                 *args,
                 **kwargs):
        super(BaseExcelView, self).__init__(*args, **kwargs)
        self.component = self.module.get_component(component)
        self.Form = Form if Form is not None else self.component.export_form

        if excel_manager is not None:
            self.excel_manager = excel_manager
        elif self.excel_manager is None:
            self.excel_manager = self.component.excel_manager

        self.EXCEL_EXPORT_RELATED = self.component.EXCEL_EXPORT_RELATED
        self.view_endpoint = view_endpoint
        self.__manager = None

    def index_url(self):
        return url_for(self.view_endpoint)

    @property
    def excel_export_actions(self):
        actions = []
        for column_set in self.EXCEL_EXPORT_RELATED:
            actions.append((url_for('.export_to_xls',
                                    related=column_set.related_attr),
                            column_set.export_label))

        return actions

    @property
    def manager(self):
        if self.__manager is None:
            self.__manager = self.excel_manager(
                self.module.managed_class, self.Form, self.EXCEL_EXPORT_RELATED)
        return self.__manager


class ExcelExport(BaseExcelView):

    def get(self):
        celery = current_app.extensions['celery']

        if celery.conf.get('CELERY_ALWAYS_EAGER', True):
            return self.generate_and_stream_file()

        return self.generate_async()

    def generate_async(self):
        task_kwargs = dict(app=self.module.crud_app.name,
                           module=self.module.id,
                           from_url=request.url,
                           user_id=current_user.id,
                           component=self.component.name,)

        if self.excel_manager != self.component.excel_manager:
            task_kwargs['manager'] = fqcn(self.excel_manager)

        task = export_task.apply_async(kwargs=task_kwargs)
        return render_template('crm/excel/async_export.html', task=task)

    def generate_and_stream_file(self):
        objects = []
        related_cs = None

        if 'import_template' not in request.args:
            objects = self.module.ordered_query(request)

        if 'related' in request.args:
            related = request.args['related']
            related_cs = ifilter(lambda cs: cs.related_attr == related,
                                 self.EXCEL_EXPORT_RELATED)
            try:
                related_cs = next(related_cs)
            except StopIteration:
                related_cs = None

        # create manager now: inside 'response_generator' we cannot instantiate the
        # manager: flask raises "RuntimeError('working outside of application
        # context')"
        manager = self.manager

        @capture_stream_errors(logger, 'Error during XLS export')
        def response_generator():
            yield ''  # start response stream before XLS build has started. For long
            # files this avoids having downstream http server returning proxy
            # error to client. Unfortunatly xlwt doesn't allow writing by
            # chunks
            workbook = manager.export(objects, related_cs)
            fd = StringIO.StringIO()
            workbook.save(fd)
            yield fd.getvalue()

        debug = request.args.get('debug_sql')
        if debug:
            # useful only in DEBUG mode, to get the debug toolbar in browser
            return '<html><body>Exported</body></html>'

        response = current_app.response_class(response_generator(),
                                              mimetype=XLSX_MIME,)
        filename = u"{}-{}.xlsx".format(self.module.managed_class.__name__,
                                        strftime("%d:%m:%Y-%H:%M:%S", gmtime()))
        response.headers['content-disposition'] = \
            'attachment;filename="{}"'.format(filename)

        return response


class TaskStatusView(views.JSONView):

    def data(self, *args, **kwargs):
        task_id = request.args.get('task_id')
        task = celery.result.AsyncResult(task_id)
        result = dict(state=task.state, exported=0, total=0)

        if task.state in ('REVOKED', 'PENDING', 'STARTED',):
            return result

        if task.state == 'PROGRESS':
            result.update(task.result)
            return result

        if task.state == 'FAILURE':
            result['message'] = _(u'An error happened during generation of file.')
            return result

        if task.state == 'SUCCESS':
            result.update(task.result)
            handle = result['handle']
            uploads = current_app.extensions['uploads']
            filemeta = uploads.get_metadata(current_user, handle)
            result['filename'] = filemeta.get('filename', 'export.xlsx')
            result['downloadUrl'] = url_for('uploads.handle',
                                            handle=handle,
                                            _external=True)
            return result

        # unattended state, return data anyway
        #FIXME: log at error level for sentry?
        return result


bp.route('/export/task_status')(TaskStatusView.as_view('task_status'))


class ExcelImport(BaseExcelView):
    methods = ['POST']

    @csrf.protect
    def post(self):
        xls = request.files['file']
        try:
            xls.stream.seek(0, 2)
            size = xls.stream.tell()
            xls.stream.seek(0)
        except:
            size = 0

        if size == 0:
            flash('Import Excel: aucun fichier fourni', 'error')
            return self.redirect_to_index()

        @capture_stream_errors(logger, 'Error during XLS import')
        def generate():
            manager = self.manager
            filename = xls.filename
            modified_items = None
            error = False
            redirect_to = url_for('.list_view')

            try:
                modified_items = manager.import_data(
                    xls, self.module.EXCEL_EXPORT_RELATED)
            # except xlrd.XLRDError as e:
            #     error = True
            #     flash(_(u'Cannot read file {filename} as Excel file').format(
            #             filename=filename),
            #           'error')
            #     logger.error(e, exc_info=True)
            except Exception, e:
                error = True
                flash(e.message, 'error')

            if modified_items is not None and len(modified_items) == 0:
                flash(_(u'No change detected in file {filename}'.format(
                    filename=filename)),
                    u'info')

            yield render_template('crm/import_xls.html',
                                  is_error=error,
                                  redirect_to=redirect_to,
                                  modified_items=modified_items,
                                  excel=manager,
                                  filename=filename)

        response = current_app.response_class(generate())
        return response


class ExcelImportValidate(BaseExcelView):
    methods = ['POST']

    @csrf.protect
    def post(self):
        action = request.form.get('_action')

        if action != u'validate':
            flash(u'Annulé', 'info')
            return self.redirect_to_index()

        filename = request.form.get('filename')
        redirect_to = url_for('.list_view')

        @capture_stream_errors(logger, 'Error during XLS import validation')
        def generate():
            # build data from form values
            f = request.form
            data = []
            item_count = int(f.get('item_count'))

            for idx in range(1, item_count + 1):
                key = 'item_{:d}_{{}}'.format(idx)
                item_id = f.get(key.format('id'))
                if item_id is not None:
                    item_id = int(item_id)
                attrs = f.getlist(key.format('attrs'))
                sig = f.get(key.format('attrs_sig'))
                to_import = frozenset(f.getlist(key.format('import')))

                # fetch object attrs
                attrkey = key.format('attr') + '_{}'
                item_modified = {}
                for attr in attrs:
                    if attr not in to_import:
                        logger.debug('item %d: skip %s', idx, attr)
                        continue
                    value = f.get(attrkey.format(attr))
                    item_modified[attr] = value

                # fetch 'many relateds' values'
                many_relateds_attrs = f.getlist(key.format(
                    'many_relateds_attrs'))
                many_relateds = {}

                for rel_attr in many_relateds_attrs:
                    rkey = key.format(rel_attr) + '_{}'
                    objs = []
                    for ridx in range(1, int(f.get(
                            rkey.format('count'), 0)) + 1):
                        modified = {}
                        robjkey = rkey.format(ridx) + '_{}'
                        r_attrs = f.getlist(robjkey.format('attrs'))

                        rattrkey = robjkey.format('attr_{}')
                        for attr in r_attrs:
                            modified[attr] = f.get(rattrkey.format(attr))

                        objs.append(_ItemUpdate(None, r_attrs, '', modified))
                    if objs:
                        many_relateds[rel_attr] = objs

                if many_relateds:
                    item_modified['__many_related__'] = many_relateds

                data.append(_ItemUpdate(item_id, attrs, sig, item_modified))

            result = self.manager.save_data(data)
            flash(_(u'Import from {filename}: {changed} items changed, '
                    '{created} items created, '
                    '{skipped} ignored due to errors').format(
                      filename=filename,
                      changed=result['changed_items'],
                      created=result['created_items'],
                      skipped=result['skipped_items']),
                  'error' if result['error_happened'] else 'info'
              )

            yield render_template('crm/xls_data_saved.html',
                                  redirect_to=redirect_to)

        response = current_app.response_class(generate())
        return response


class ExcelModuleComponent(ModuleComponent):
    """
    A :class:`ModuleComponent <Component>` for
    :class:`abilian.web.frontend.Module` objects
    """
    name = 'excel'
    EXCEL_SUPPORT_IMPORT = False

    #: tuple of ManyRelatedColumnSet()
    EXCEL_EXPORT_RELATED = ()

    excel_manager = ExcelManager
    export_form = None

    def __init__(self, export_form=None, excel_manager=None, *args, **kwargs):
        super(ExcelModuleComponent, self).__init__(*args, **kwargs)

        if export_form is not None:
            self.export_form = export_form
        if excel_manager is not None:
            self.excel_manager = excel_manager

    def init(self, *args, **kwargs):
        super(ExcelModuleComponent, self).init(*args, **kwargs)
        module = self.module
        endpoint = module.endpoint

        if self.export_form is None:
            self.export_form = module.edit_form_class

        module._setup_view('/export_xls',
                           b'export_xls',
                           ExcelExport,
                           module=module,
                           component=self.name,
                           excel_manager=self.excel_manager,
                           Form=self.export_form,
                           view_endpoint=endpoint + '.list_view')

        if not self.EXCEL_SUPPORT_IMPORT:
            return

        module._setup_view('/validate_imported_data',
                           b'validate_imported_xls',
                           ExcelImportValidate,
                           methods=['POST'],
                           component=self.name,
                           module=module,
                           view_endpoint=endpoint + '.list_view')

        module._setup_view('/import_xls',
                           b'import_xls',
                           ExcelImport,
                           methods=['POST'],
                           component=self.name,
                           module=module,
                           view_endpoint=endpoint + '.list_view')

    def get_actions(self):
        excel_actions = []
        button = 'default' if not self.EXCEL_EXPORT_RELATED else None
        endpoint = self.module.endpoint
        excel_actions.append(ModuleAction(self.module,
                                          'excel',
                                          'export_xls',
                                          title=_l(u'Export to Excel'),
                                          icon=FAIcon('align-justify'),
                                          endpoint=Endpoint(endpoint +
                                                            '.export_xls'),
                                          button=button,
                                          css='datatable-export'))

        for column_set in self.EXCEL_EXPORT_RELATED:
            excel_actions.append(ModuleActionGroupItem(
                self.module,
                'excel',
                'export_related_' + column_set.related_attr,
                title=column_set.export_label,
                icon=FAIcon('align-justify'),
                css='datatable-export',
                endpoint=Endpoint(endpoint + '.export_xls',
                                  related=column_set.related_attr)))

        if self.EXCEL_SUPPORT_IMPORT:
            pass

        if len(excel_actions) > 1:
            excel_actions = [ModuleActionDropDown(self.module,
                                                  'excel',
                                                  'actions',
                                                  title=_l(u'Excel'),
                                                  button='default',
                                                  items=excel_actions)]

        return excel_actions



# coding=utf-8
"""
"""
from __future__ import absolute_import

from .related import ManyRelatedColumnSet


class ManyYearlyColumnSet(ManyRelatedColumnSet):

    def __init__(self, related_attr='ignored', *args, **kwargs):
        super(ManyYearlyColumnSet, self).__init__(
            related_attr='__yearly_data__',
            *args,
            **kwargs)

    def iter_items(self, obj):
        for data in getattr(obj, self.related_attr).values():
            yield data



# coding=utf-8
"""
"""
from __future__ import absolute_import

from wtforms.fields import IntegerField, StringField, TextAreaField
from wtforms.widgets import HiddenInput

from abilian.i18n import _l, country_choices
from abilian.web.forms import ModelForm
from abilian.web.forms.fields import ModelFormField, Select2Field
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import flaghidden, optional, required
from abilian.web.forms.widgets import ModelWidget, TextArea

from .models import PhoneNumber, PostalAddress
from .widgets import PhoneNumberWidget


class RequireableFormField(object):
    """
  Mixin for Formfield based class, to allow toggle required / optional

  Basic FormField doesn't allow any validators, making harder to include an
  optional formfield with required fields.
  """

    def __init__(self, *args, **kwargs):
        self.__validators = tuple(kwargs.pop('validators', ()))
        super(RequireableFormField, self).__init__(*args, **kwargs)

        for v in self.__validators:
            for flag in getattr(v, 'field_flags', ()):
                if flag == 'required':
                    self.flags.required = True
                elif flag != 'optional':
                    raise TypeError('{} accept only "required" validator'
                                    ''.format(self.__class__.__name__))

    def process(self, *args, **kwargs):
        super(RequireableFormField, self).process(*args, **kwargs)

        for f in self.form:
            if f.flags.required:
                f.validators = self.__filter_validators(f)
                f.flags.required = self.flags.required
                f.flags.optional = not self.flags.required

    def __filter_validators(self, field):
        required = self.flags.required
        if required == field.flags.required:
            return field.validators

        validators = []
        remove = 'optional' if required else 'required'
        has_required = False

        for v in field.validators:
            if required and any(f == 'required' for f in v.field_flags):
                has_required = True
            if any(f == remove for f in v.field_flags):
                continue
            validators.append(v)

        if required and not has_required:
            validators.append(required())

        return validators


class PhoneNumberField(StringField):
    widget = PhoneNumberWidget()


class PostalAddressForm(ModelForm):

    id = IntegerField(widget=HiddenInput(),
                      validators=[optional(), flaghidden()])
    street_lines = TextAreaField(
        _l(u'postal_address_street_lines'),
        description=_l(u'postal_address_street_lines_help'),
        validators=[required()],
        filters=(strip,),
        widget=TextArea(rows=4,
                        resizeable=None),)

    administrative_area = StringField(
        _l(u'postal_address_administrative_area'),
        description=_l(u'postal_address_administrative_area_help'),)

    sub_administrative_area = StringField(
        _l(u'postal_address_sub_administrative_area'),
        description=_l(u'postal_address_sub_administrative_area_help'),)

    postal_code = StringField(
        _l(u'postal_address_postal_code'),
        description=_l(u'postal_address_postal_code_help'),
        validators=[required()],
        filters=(strip,),)

    locality = StringField(
        _l(u'postal_address_locality'),
        # description=_l(u'postal_address_locality_help'),
        validators=[required()],
        filters=(strip,),)

    country = Select2Field(
        _l(u'postal_address_country'),
        validators=[required()],
        filters=(strip,),
        choices=country_choices,)

    class Meta:
        model = PostalAddress
        include_primary_keys = True
        assign_required = False


class PostalAddressField(RequireableFormField, ModelFormField):
    widget = ModelWidget()

    def __init__(self, *args, **kwargs):
        super(PostalAddressField, self).__init__(PostalAddressForm, *args,
                                                 **kwargs)


class PhoneNumberForm(ModelForm):
    id = IntegerField(widget=HiddenInput(),
                      validators=[optional(), flaghidden()])
    type = StringField(
        _l(u'phonenumber_type'),
        description=_l(u'phonenumber_type_help'),)
    number = PhoneNumberField(
        _l(u'phonenumber_number'),
        description=_l(u'for an extension number add "#1234"'),
        validators=[required()])

    class Meta:
        model = PhoneNumber
        include_primary_keys = True
        assign_required = False


class PhoneNumberFormField(RequireableFormField, ModelFormField):
    widget = ModelWidget(
        view_template='crm/widgets/phonenumber_model_view.html')

    def __init__(self, *args, **kwargs):
        super(PhoneNumberFormField, self).__init__(PhoneNumberForm, *args,
                                                   **kwargs)



# coding=utf-8
"""
"""
from __future__ import absolute_import


class ExcelError(Exception):
    pass


class ExcelImportError(ExcelError, ValueError):

    def __init__(self, message, imported_value, *args):
        ValueError.__init__(self, message, *args)
        self.imported_value = imported_value



# coding=utf-8
"""
"""
from __future__ import absolute_import

import re
from collections import OrderedDict

import sqlalchemy as sa
import wtforms.fields
from future.utils import string_types

import abilian.web.forms.fields as awbff
import abilian.web.forms.validators as aw_validators
import abilian.web.forms.widgets as aw_widgets

from ..definitions import (FORM_FILTERS, LIST_GENERATORS, MAX_IDENTIFIER_LENGTH,
                           VALIDATORS, WIDGETS)
from .registry import Registrable, get_formfield

_VALID_IDENTIFIER_RE = re.compile(r'[A-Za-z_][A-Za-z0-9_]*', re.UNICODE)


def assert_valid_identifier(s):
    match = _VALID_IDENTIFIER_RE.match(s)
    if match is None or match.end() != match.endpos:
        raise ValueError('{} is not a valid python identifier'.format(repr(s)))


class Field(Registrable):

    name = None
    label = u''

    #: sqlalchemy column type
    sa_type = None

    #: default form field type
    default_ff_type = 'TextField'

    #: set to False is field supports only single values
    allow_multiple = True

    def __init__(self, model, data, generator):
        """
    """
        self.model = model
        self.data = data
        self.generator = generator
        self.name = data['name']
        assert_valid_identifier(self.name)
        self.label = data.get('description', u'')
        self.sa_type_options = data.get('type_options', dict())
        self.required = data.get('required', False)
        self.unique = data.get('unique', False)
        self.indexed = data.get('indexed', False)
        self.multiple = data.get('multiple', False)

        if self.multiple and not self.allow_multiple:
            raise ValueError("Field {!r}: {!r} doesn't support multiple values"
                             "".format(self.name, self.__class__.__fieldname__))

        self.nullable = data.get('nullable', not self.required)
        if self.nullable and self.required:
            # if required, then it cannot be null
            self.nullable = False

        ff_type = self.get_ff_type()
        data['formfield'] = ff_type(model=model, data=data, generator=generator)

    def get_ff_type(self, *args, **kwargs):
        ff_type = self.default_ff_type
        if not isinstance(ff_type, FormField):
            ff_type = get_formfield(ff_type)
        return ff_type

    def get_model_attributes(self, *args, **kwargs):
        """
    Return attributes to be set on generated model.

    :return: iterable of `(name, type)`
    """
        col_name = self.name[:MAX_IDENTIFIER_LENGTH]
        extra_args = {'nullable': self.nullable}
        extra_args['info'] = info = {}
        info['label'] = self.label

        if 'default' in self.data:
            extra_args['default'] = self.data.get('default')

        if self.indexed:
            info['searchable'] = True
            info['index_to'] = ('text',)

        if 'from_list' in self.data:
            info['choices'] = OrderedDict(self.data['from_list'])

        attr = sa.schema.Column(col_name, self.sa_type(**self.sa_type_options),
                                **extra_args)

        return ((self.name, attr),)

    def get_table_args(self, *args, **kwargs):
        """
    Arguments to be added to __table_args__

    :return: iterable
    """
        args = []
        if self.unique:
            args.append(sa.UniqueConstraint(self.name[:MAX_IDENTIFIER_LENGTH]))

        return args

    def get_field(self, *args, **kwargs):
        """
    return a tuple (attribute name, `wtforms.fields.Field` instance)
    """
        field_kw = self.get_field_extra_args(self, *args, **kwargs)
        field_type = self.get_field_type()
        field = field_type(self.description, field_kw)
        return self.name, field


class FormField(Registrable):
    """
  name: Text
  help: Text
  lines: int   -> TextAreaField
  from_list:      -> Select2Field
  from_function:  -> Select2Field
  type_options: Dict
  description: Text
  multiple: Boolean
  required: Boolean
  validator_length_max: int > 0
  validator_length_min: int > 0
  """
    #: form field type
    ff_type = wtforms.fields.TextField

    def __init__(self, model, data, generator):
        self.model = model
        self.data = data
        self.generator = generator
        self.name = data['name']
        assert_valid_identifier(self.name)
        self.label = data.get('description', u'')
        self.sa_type_options = data.get('type_options', dict())
        self.required = data.get('required', False)
        self.multiple = data.get('multiple', False)
        self.validator_length_max = data.get('validator_length_max', -1)
        self.validator_length_min = data.get('validator_length_min', -1)
        if self.validator_length_max < -1:
            self.validator_length_max = -1
        if self.validator_length_min < -1:
            self.validator_length_min = -1

    def get_form_attributes(self):
        field_type = self.get_type()
        extra_args = self.get_extra_args()
        yield self.name, field_type(label=self.label, **extra_args)

    def get_type(self, *args, **kwargs):
        field_type = self.ff_type

        if 'lines' in self.data:
            field_type = wtforms.fields.TextAreaField

        if 'from_list' in self.data or 'from_function' in self.data:
            field_type = (awbff.Select2Field if not self.multiple else
                          awbff.Select2MultipleField)
        return field_type

    def get_extra_args(self, *args, **kwargs):
        extra_args = {
            'filters': list(self.get_filters()),
            'validators': self.get_validators(),
        }

        # extra validators & filters specified in data
        d = self.data

        description = d.get('help', u'').strip()
        if description:
            extra_args['description'] = description

        if 'validators' in d:
            validators = d['validators']
            if isinstance(validators, string_types):
                validators = [validators]
            validators = [VALIDATORS[v]() for v in validators]
            extra_args['validators'].extend(validators)

        if 'filters' in d:
            filters = d['filters']
            if isinstance(filters, string_types):
                filters = [filters]
            filters = [FORM_FILTERS[f] for f in filters]
            extra_args['filters'].extend(filters)

        for key in ('filters', 'validators',):
            if not extra_args[key]:
                del extra_args[key]

        self.setup_widgets(extra_args)
        return extra_args

    def get_filters(self, *args, **kwargs):
        """
    Default filters
    """
        return ()

    def get_validators(self, *args, **kwargs):
        """
    Default validators
    """
        validators = []
        if self.required:
            validators.append(aw_validators.required())
        else:
            validators.append(aw_validators.optional())

        if self.validator_length_max != -1 or self.validator_length_min != -1:
            validators.append(
                aw_validators.Length(min=self.validator_length_min,
                                     max=self.validator_length_max))
        return validators

    def setup_widgets(self, extra_args):
        """
    set 'widget' and 'view_widget' in extra_args
    """
        d = self.data

        if 'from_list' in d:
            if self.multiple:
                extra_args['view_widget'] = aw_widgets.ListWidget()

            options = list(d['from_list'])
            if 'required' in d:
                if options[0][0] == u'':
                    options.pop(0)
            elif options[0] != u'':
                options.insert(0, (u'', u''))

            extra_args['choices'] = options

        if 'from_function' in d:
            if self.multiple:
                extra_args['view_widget'] = aw_widgets.ListWidget()
            if 'from_function_nocall' in d and d['from_function_nocall']:
                extra_args['choices'] = LIST_GENERATORS[d['from_function']]
            else:
                extra_args['choices'] = LIST_GENERATORS[d['from_function']]()

        if 'lines' in d:
            extra_args['widget'] = aw_widgets.TextArea(resizeable='vertical',
                                                       rows=d['lines'])

        self.setup_widgets_from_data(extra_args)

    def setup_widgets_from_data(self, extra_args):
        """
    Setup widgets from specification in data ('widget' and 'view_widget' entries)
    """
        d = self.data
        for widget_arg in ('widget', 'view_widget'):
            if widget_arg not in d:
                continue

            widget = d[widget_arg]

            if isinstance(widget, (str, unicode)):
                if widget not in WIDGETS:
                    raise ValueError('Invalid {}: {}'.format(
                        widget_arg, widget.encode('utf-8')))
                widget = WIDGETS[widget]
                kw = d.get(widget_arg + '_args', dict())
                widget = widget(**kw)

            extra_args[widget_arg] = widget



# coding=utf-8
"""
"""
from __future__ import absolute_import

import datetime
import hashlib
import logging
import math
import pprint
from collections import defaultdict
from functools import partial
from itertools import chain, islice, izip
from operator import attrgetter

import itsdangerous
import openpyxl
import sqlalchemy as sa
from flask import current_app
from openpyxl import Workbook, styles
from openpyxl.cell import STRING_TYPES
from openpyxl.utils import get_column_letter, units
from openpyxl.writer.write_only import WriteOnlyCell

from abilian.core.extensions import db
from abilian.core.sqlalchemy import JSON as JSONType
from abilian.i18n import _
from abilian.services.vocabularies.models import BaseVocabulary
from abilian.web.forms.fields import ModelFieldList

from ..models import PostalAddress
from .columns import (Column, ColumnSet, DateColumn, DateTimeColumn, Invalid,
                      ManyRelatedColumnSet, PostalAddressColumn,
                      RelatedColumnSet, VocabularyColumn)
from .exc import ExcelError, ExcelImportError

logger = logging.getLogger(__name__)

#: column minimum width
MIN_WIDTH = 3

#: column maximum width
MAX_WIDTH = units.BASE_COL_WIDTH * 2


class ExcelManager(object):
    """
    import/export data to excel files, from/to model
    """

    XF_HEADER = {
        'font': styles.Font(bold=True),
        'alignment': styles.Alignment(horizontal='center',
                                      vertical='top',
                                      wrapText=True)
    }
    XF_EDITABLE = styles.Protection(locked=False)
    XF_DATE_EDITABLE = styles.Protection(locked=False)
    XF_DATE_NUMBER_FORMAT = "DD/MM/YYYY"

    #: column to identify item by (unique) name (if no database id is
    # available). This is used by importer to import a many related sheet
    ID_BY_NAME_COL = None

    #: list of columns that identify a unique object. If an import sheet has no id
    # one of this columns may identify an object. In this case the object is
    # updated with values from non-empty cells (like a dict.update())
    UNIQUE_ID_COLS = ()

    #: sheet name where data should be, at import or export
    MAIN_SHEET_NAME = u'Sheet 1'

    #: For many related export/import, set column at which object's columns is
    # splitted to insert related object columns
    MANY_SPLIT_COLUMN = 2

    #: list of field to skip
    SKIP_COLS = ()

    def __init__(self, model_cls, form_cls, many_related_cs=()):
        """
    model_cls: Model class
    form_cls: Form class
    many_related_cs: iterable of ManyRelatedColumnSet instances
    """
        self.MANY_RELATED_CS = many_related_cs
        self.model_cls = model_cls
        self.form_cls = form_cls
        self.form = form_cls(csrf_enabled=False)
        self.mapper = sa.inspect(model_cls)
        self.db_columns = self.mapper.c

        if self.ID_BY_NAME_COL and self.ID_BY_NAME_COL not in self.UNIQUE_ID_COLS:
            self.UNIQUE_ID_COLS += (self.ID_BY_NAME_COL,)

        # collect exportable attributes
        columns = []

        if 'id' not in self.form and 'id' not in self.SKIP_COLS:
            columns.append(Column('id', 'id', int))

        for field in self.form:
            info = self.columns_for(field, self.form)
            if info is not None:
                columns.append(info)

        columns.extend(self.additional_columns())
        self.columns = ColumnSet(*columns)
        # logger.debug('\nColumns:\n %s', pprint.pformat([c for c in self.columns]))

    def additional_columns(self):
        """
    To be overriden by subclasses that want to add specific columns
    """
        return []

    _signer = None

    @property
    def signer(self):
        if self._signer is None:
            # signer is created on 1st access. When it's never accessed, this instance
            # can be used outside a flask application context
            config = current_app.config
            secret_key = config.get('SECRET_KEY')
            self._signer = itsdangerous.TimestampSigner(secret_key,
                                                        salt=__name__)

        return self._signer

    @property
    def attrs(self):
        return self.columns.attrs

    def attrs_signature(self, columns=None):
        """
    Return hash for current labels.
    """
        if columns is None:
            columns = self.columns
        md5 = hashlib.md5()
        for attr in columns.attrs:
            self.update_md5(md5, attr)
        return md5.hexdigest()

    @property
    def labels(self):
        return self.columns.labels

    def get_attr_to_column(self, columnset, map_related_attr=False):
        """
    map attribute names to their respective Column instance.

    If map_related_attr is True, also map related's attributes to their
    RelatedColumnSet instance
    """
        attr_map = dict(izip(columnset.attrs, columnset.iter_flatened()))
        # we also want to collect related
        # attributes of RelatedColumnSets. self.columns.attrs doesn't show them
        for col in columnset:
            if isinstance(col, RelatedColumnSet):
                attr_map[col.related_attr] = col
                if map_related_attr:
                    for attr in col.attrs:
                        attr_map[attr] = col

        return attr_map

    @property
    def attr_to_column(self):
        if not hasattr(self, '_attr_to_column'):
            self._attr_to_column = self.get_attr_to_column(self.columns)
        return self._attr_to_column

    @property
    def attr_to_main_column(self):
        """ Return a mapping of attribute name: column instance, with the exception
    of attributes of related instances mapped to a RelatedColumnSet() instance.
    """
        if not hasattr(self, '_attr_to_main_column'):
            self._attr_to_main_column = self.get_attr_to_column(
                self.columns,
                map_related_attr=True)
        return self._attr_to_main_column

    def export(self, objects, related_column_set=None, progress_callback=None):
        """
    Exports objects to a Workbook
    """
        if related_column_set is not None:
            return self.export_many(objects,
                                    related_column_set,
                                    progress_callback=progress_callback)

        wb = Workbook()
        if wb.worksheets:
            wb.remove_sheet(wb.active)

        ws, row = self._new_export_sheet(wb, self.model_cls.__name__,
                                         self.columns)

        cols_width = [MIN_WIDTH]
        for idx, label in enumerate(self.columns.labels, 2):
            letter = get_column_letter(idx)
            cols_width.append(ws.column_dimensions[letter].width)

        total = objects.count() if hasattr(objects, 'count') else len(objects)
        exported = 0
        if progress_callback:
            progress_callback(exported=exported, total=total)

        for r, obj in enumerate(objects, row):
            md5 = hashlib.md5()
            offset = 0
            cells = [WriteOnlyCell(ws)]
            for c, col in enumerate(self.columns, 1):
                # one model column might be exported in multiple excel columns. Example:
                # a related entity like a contact: fullname, email, etc
                for rel_idx, (import_val, value) in enumerate(col.data(obj)):
                    cell = WriteOnlyCell(ws, value=import_val)
                    self.style_for(cell)
                    cells.append(cell)
                    self.update_md5(md5, import_val)

                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                offset += rel_idx

            cells[0].value = self.signer.sign(md5.hexdigest())
            ws.append(cells)

            exported += 1
            if progress_callback:
                progress_callback(exported=exported, total=total)

        if progress_callback:
            progress_callback(exported=total, total=total)

        self.finalize_worksheet(ws, cols_width=cols_width)
        return wb

    def export_many(self, objects, related_columns_set, progress_callback=None):
        """
    :param related_columns_set: a :class:`ManyRelatedColumnSet` instance
    """
        assert isinstance(related_columns_set, ManyRelatedColumnSet)

        wb = Workbook()
        if wb.worksheets:
            wb.remove_sheet(wb.active)

        total = objects.count() if hasattr(objects, 'count') else len(objects)
        exported = 0
        if progress_callback:
            progress_callback(exported=exported, total=total)

        all_columns = self._columns_for_many_related(related_columns_set)
        ws, start_row = self._new_export_sheet(
            wb, related_columns_set.export_label, all_columns)
        related_columns_len = related_columns_set.colspan

        cols_width = [MIN_WIDTH]
        for idx, label in enumerate(all_columns.labels, 2):
            letter = get_column_letter(idx)
            cols_width.append(ws.column_dimensions[letter].width)

        row_offset = 0
        for r, obj in enumerate(objects, start_row):
            row_offset -= 1
            # prepare main item data: it will be repeated for each of its related
            # objects
            columns = iter(self.columns)

            head_data = [import_val
                         for col in islice(columns, self.MANY_SPLIT_COLUMN)
                         for import_val, value in col.data(obj)]

            # islice has started to consume 'columns' iterator, iter remaining columns
            tail_data = [import_val
                         for col in columns
                         for import_val, value in col.data(obj)]

            # generate rows for each related object
            is_empty = True
            md5 = hashlib.md5()
            for item in related_columns_set.iter_items(obj):
                cells = [WriteOnlyCell(ws)]
                is_empty = False
                row_offset += 1
                col_offset = 1
                for c, val in enumerate(head_data, 1):
                    # FIXME: we could keep HEAD_MD5, and just extends `cells` with
                    # head_data
                    cell = WriteOnlyCell(ws, value=val)
                    self.style_for(cell)
                    cells.append(cell)
                    self.update_md5(md5, value)
                    # ws.write(r+row_offset, c, value, style)
                    col_offset += 1

                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                data = related_columns_set.data(item)
                for c, (import_val, value) in enumerate(data, col_offset):
                    cell = WriteOnlyCell(ws, value=import_val)
                    self.style_for(cell)
                    cells.append(cell)
                    #ws.write(r+row_offset, c, import_val, self.style_for(import_val))
                    self.update_md5(md5, import_val)
                    col_offset += 1

                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                for c, val in enumerate(tail_data, col_offset):
                    cell = WriteOnlyCell(ws, value=val)
                    self.style_for(cell)
                    cells.append(cell)
                    #ws.write(r+row_offset, c, value, style)
                    self.update_md5(md5, value)

                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                cells[0].value = self.signer.sign(md5.hexdigest())
                ws.append(cells)

            if is_empty:
                # no related items. We still need to output 1 line, with empty columns
                # for the related part.
                cells = [WriteOnlyCell(ws)]
                row_offset += 1
                col_offset = 1
                # head data 1st, then adjust col_offset to skip related item columns,
                # then tail data
                for c, val in enumerate(head_data, 1):
                    cell = WriteOnlyCell(ws, value=val)
                    self.style_for(cell)
                    cells.append(cell)
                    self.update_md5(md5, value)
                    col_offset += 1
                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                for ignored in range(related_columns_len):
                    cells.append(WriteOnlyCell(ws))

                col_offset += related_columns_len

                for c, val in enumerate(tail_data, col_offset):
                    cell = WriteOnlyCell(ws, value=val)
                    self.style_for(cell)
                    cells.append(cell)
                    self.update_md5(md5, value)
                    # estimate width
                    value = unicode(cell.value)
                    width = max(len(l) for l in value.split(u'\n')) + 1
                    cols_width[c] = max(width, cols_width[c])

                cells[0].value = self.signer.sign(md5.hexdigest())
                ws.append(cells)

            exported += 1
            if progress_callback:
                progress_callback(exported=exported, total=total)

        if progress_callback:
            progress_callback(exported=total, total=total)

        self.finalize_worksheet(ws, cols_width=cols_width)
        return wb

    def _new_export_sheet(self, wb, name, columns):
        ws = wb.create_sheet(title=name)
        #ws.protect = True

        # attributes row
        row = 0
        md5 = hashlib.md5()
        cells = [WriteOnlyCell(ws)]
        for c, attr in enumerate(columns.attrs, 1):
            cells.append(WriteOnlyCell(ws, value=attr))
            self.update_md5(md5, attr)

        cells[0].value = self.signer.sign(md5.hexdigest())
        ws.append(cells)
        row += 1

        #  labels row
        md5 = hashlib.md5()
        cells = [WriteOnlyCell(ws)]
        for c, label in enumerate(columns.labels, 1):
            cell = WriteOnlyCell(ws, value=unicode(label))
            cell.font = self.XF_HEADER['font']
            cell.alignment = self.XF_HEADER['alignment']
            cells.append(cell)
            self.update_md5(md5, label)

        cells[0].value = self.signer.sign(md5.hexdigest())
        ws.append(cells)
        row += 1

        # row / column properties
        ws.row_dimensions[1].hidden = True  # hide attributes names row
        ws.column_dimensions['A'].hidden = True  # hide md5 column
        overflow = 0

        for idx, cell in enumerate(cells, 1):
            letter = get_column_letter(idx)
            width = max(len(l) for l in cell.value.split(u'\n')) + 1
            if width > MAX_WIDTH:
                overflow = max(overflow, width)
                width = MAX_WIDTH
            # MIN_WIDTH <= custom width <= BASE_COL_WIDTH * 2
            ws.column_dimensions[letter].width = max(MIN_WIDTH, width)

        if overflow:
            lines = int(math.ceil(overflow / float(MAX_WIDTH)))
            ws.row_dimensions[2].height = units.DEFAULT_ROW_HEIGHT * lines

        return ws, row

    def finalize_worksheet(self, ws, cols_width=None):
        # if we do this during initialize, next ws.append() will occur at line 4,
        # leaving a blank line.
        ws.freeze_panes = ws.cell(row=3, column=3)

        if cols_width is not None:
            for idx, width in enumerate(cols_width, 1):
                letter = get_column_letter(idx)
                width = min(max(width, MIN_WIDTH), MAX_WIDTH)
                ws.column_dimensions[letter].width = width

    def _columns_for_many_related(self, related_cs):
        """
    Given a ManyRelatedColumnSet instance, returns a columns set with object
    columns and related columns
    """
        SPLIT_COLUMN = self.MANY_SPLIT_COLUMN
        main_cols = list(self.columns)
        head_cols = main_cols[:SPLIT_COLUMN]
        tail_cols = main_cols[SPLIT_COLUMN:]
        all_columns = ColumnSet(
            *list(chain(head_cols, (related_cs,), tail_cols)))
        return all_columns

    def style_for(self, cell):
        """ Return XLWT style to use for value. """
        cell.protection = self.XF_EDITABLE
        if isinstance(cell.value, datetime.date):
            cell.protection = self.XF_DATE_EDITABLE
            cell.number_format = self.XF_DATE_NUMBER_FORMAT

    def import_data(self, xls_file, many_related_columns_set=()):
        """ Read xls file, detect changed rows; then for each find changed
        attributes of related item.

        return a list of rows; each row is a dict with:
          - `item`: db object
          - `modified`: dict attr: (current value, new value)
          - `attr_sig`: signature of 'modified', required for 'save_data()'
        """
        wb = openpyxl.load_workbook(xls_file)
        try:
            ws = wb[self.model_cls.__name__]
        except KeyError:
            # last chance
            ws = wb['Sheet 1']

        self._validate_import_sheet(ws)
        data_rows = self._collect_changed_rows(ws, wb)
        by_id = {}
        by_name = {}
        unique_id_cols = {}
        for d in data_rows:
            item_id = d.get('id')
            if item_id and not isinstance(item_id, Invalid):
                by_id[item_id] = d
            name = d['__metadata__'].get('unique_name')
            if name is not None:
                # FIXME: check name is unique
                by_name[name] = d
            uid_cols = d['__metadata__'].get('unique_id_cols')
            if uid_cols is not None:
                for key, val in uid_cols.iteritems():
                    unique_id_cols.setdefault(key, {})[val] = d

        # collect data from the 'many related' sets
        for related_cs in many_related_columns_set:
            self._collect_changed_relateds(related_cs, wb, data_rows, by_id,
                                           by_name)

        # prefetch items
        ids = [d['id'] for d in data_rows if 'id' in d]
        q = self.model_cls.query
        q.filter(self.model_cls.id.in_(ids)).all()

        # detect modified attributes
        attr_to_col = self.attr_to_column
        modified_items = []

        for data in data_rows:
            item = None
            metadata = data.pop('__metadata__')
            is_new = metadata.get('is_new')
            is_update = metadata.get('is_update')

            if not is_new:
                try:
                    item = q.get(data['id'])
                except:
                    # this may be a deleted line from an exported file, but excel has
                    # preserved hidden cell
                    continue
                if item is None:
                    continue
            else:
                # new item, or update sheet: incomplete lines of existing items
                item = self.model_cls()

                for col_name in unique_id_cols:
                    val = data.get(col_name)
                    if val is not None:
                        try:
                            item = q.filter(getattr(self.model_cls, col_name) ==
                                            val).one()
                        except sa.orm.exc.NoResultFound:
                            pass
                        except sa.orm.exc.MultipleResultsFound:
                            # log this to sentry
                            logger.error('Multiple items match %s[%s] == ""%s"',
                                         self.model_cls.__name__,
                                         col_name,
                                         unicode(val).encode('utf-8'),
                                         extra={'stack': True})
                        else:
                            is_new = metadata['is_new'] = False
                            is_update = metadata['is_update'] = True
                            data['id'] = item.id

            assert not is_update or (is_update and not is_new)

            many_related_data = {}
            for cs in self.MANY_RELATED_CS:
                many_related_data[cs.related_attr] = data.pop(cs.related_attr,
                                                              None)

            modified = self._get_modified_data(item, data, attr_to_col, is_new,
                                               is_update)
            required_missing = modified.pop('__required_missing__', False)

            modified_relateds = {}
            for cs in self.MANY_RELATED_CS:
                related_data = many_related_data.get(cs.related_attr)
                if not related_data:
                    continue

                rel_items = []
                name_getter = attrgetter(cs.ID_BY_NAME_COL)
                current_data = getattr(item, cs.related_attr)
                current_data_map = {name_getter(o): o for o in current_data}
                related_attr_to_col = self.get_attr_to_column(
                    related_cs,
                    map_related_attr=True)

                for item_data in related_data:
                    rel_metadata = item_data.pop('__metadata__')
                    name = rel_metadata.get('__metadata__',
                                            {}).get('unique_name')
                    is_new = name is not None
                    default = cs.model_cls()
                    obj = current_data_map.get(name,
                                               default) if is_new else default
                    rel_modified = self._get_modified_data(
                        obj, item_data, related_attr_to_col, is_new)

                    if rel_modified:
                        rel_items.append(dict(modified=rel_modified,
                                              metadata=rel_metadata))

                if rel_items:
                    modified_relateds[cs.related_attr] = rel_items

            if modified or modified_relateds:
                # FIXME: use itsdangerous.TimedSerializer instead of joining a list
                valid_keys = (k
                              for k in modified
                              if not (required_missing or modified[k].error))
                attr_sig = self.signer.sign(u';'.join(sorted(valid_keys)))
                attr_sig = self.extract_signature(attr_sig)

                modified_items.append(dict(item=item,
                                           metadata=metadata,
                                           required_missing=required_missing,
                                           modified=modified,
                                           many_relateds=modified_relateds,
                                           attr_signature=attr_sig))

        # FIXME: also return list of invalid_rows?
        return modified_items

    def _validate_import_sheet(self, ws, columns=None):
        """
        Validates signatures of column lines: columns ordering and labels should
        not have been changed
        """
        orig_md5 = self.get_cell_value(ws, 0, 0)
        try:
            orig_md5 = self.signer.unsign(orig_md5)
        except itsdangerous.BadSignature:
            raise ExcelError(u'La signature de la 1ere ligne est incorrecte, '
                             'impossible de valider le fichier')

        if columns is None:
            columns = self.columns

        md5 = hashlib.md5()
        for col in range(1, ws.ncols):
            attr = self.get_cell_value(ws, 0, col)
            self.update_md5(md5, attr)

        if orig_md5 != md5.hexdigest():
            raise ExcelError(
                u'L\'ordre des colonnes a changé ou leurs étiquettes ont'
                u' été modifiées: le fichier n\'est pas valide.')

        if orig_md5 != self.attrs_signature(columns):
            raise ExcelError(
                u'La feuille "{}" utilise une version du format d\'export qui'
                u' n\'est plus à jour'.format(ws.name))

    def _collect_changed_rows(self, ws, wb):
        """
        Collect modified rows: compare md5@column(0) with md5(row values), and
        new rows.
        """
        attr_to_main = self.attr_to_main_column
        attr_to_col = self.attr_to_column
        single_value_attrs = {k
                              for k, column in attr_to_main.items()
                              if not isinstance(column, RelatedColumnSet)}
        invalid_lines = []
        data_rows = []

        for row in range(2, ws.nrows):
            cell = partial(self.get_cell_value, ws, row)
            data = defaultdict(dict)
            orig_md5 = cell(0)
            is_new = False if orig_md5 else True
            data['__metadata__'] = {'row': row + 1,
                                    'is_new': is_new,
                                    'is_update': False,
                                    'unique_id_cols': dict()}

            if not is_new:
                try:
                    orig_md5 = self.signer.unsign(orig_md5)
                except itsdangerous.BadSignature:
                    invalid_lines.append(row)
                    logger.debug('Skip line %d: md5 signature is invalid', row)
                    continue

            md5 = hashlib.md5()
            for idx, attr in enumerate(self.attrs, 1):
                value = cell(idx)
                if is_new and attr == 'id':
                    data['__metadata__']['id'] = value
                    continue

                value = cell(idx)
                value = self.update_import_data(
                    data, attr, value, attr_to_col, attr_to_main,
                    self.ID_BY_NAME_COL, self.UNIQUE_ID_COLS, wb)
                self.update_md5(md5, value)

            md5 = md5.hexdigest()
            if is_new or orig_md5 != md5:
                for attr in single_value_attrs:
                    if attr in data:
                        data[attr] = data[attr][attr]
                data_rows.append(data)

        return data_rows

    def _collect_changed_relateds(self, related_cs, wb, data_rows, id_map,
                                  name_map):
        """
        Collect changes from a many related column set. Data must be in sheet
        with named after `related_cs.export_label
        """
        if not self.ID_BY_NAME_COL:
            # the manager is not configured to be able find model by a name
            # column. Since main object id is not exported in 'many_related' exports
            # we cannot process imports for this model_cls.
            return

        try:
            ws = wb[related_cs.export_label]
        except KeyError:
            logger.debug('No sheet named "%s"',
                         related_cs.export_label.encode('utf-8'))
            return

        all_columns = self._columns_for_many_related(related_cs)
        try:
            self._validate_import_sheet(ws, all_columns)
        except ExcelError:
            logger.warning(
                'Many related _collect_changed_relateds: Invalid header signature, '
                'skipping (%s)', related_cs.export_label.encode('utf-8'))
            return

        attr_to_main = self.attr_to_main_column
        attr_to_col = self.attr_to_column
        related_attr_to_main = self.get_attr_to_column(related_cs, True)
        related_attr_to_col = self.get_attr_to_column(related_cs)

        single_value_attrs = {k
                              for k, column in related_attr_to_main.items()
                              if not isinstance(column, RelatedColumnSet)}
        invalid_lines = []

        for row in range(2, ws.nrows):
            cell = partial(self.get_cell_value, ws, row)
            data = defaultdict(dict)
            data['__metadata__'] = {'row': row + 1}
            related_data = defaultdict(dict)
            related_data['__metadata__'] = {'row': row + 1}

            orig_md5 = cell(0)
            is_new = False if orig_md5 else True

            if not is_new:
                try:
                    orig_md5 = self.signer.unsign(orig_md5)
                except itsdangerous.BadSignature:
                    invalid_lines.append(row)
                    logger.debug('Skip line %d: md5 signature is invalid', row)
                    continue

            attrs = iter(self.attrs)
            md5 = hashlib.md5()
            # main item: head data
            for idx, attr in enumerate(
                    islice(attrs, self.MANY_SPLIT_COLUMN), 1):
                if is_new and attr == 'id':
                    continue
                value = cell(idx)
                value = self.update_import_data(data, attr, value, attr_to_col,
                                                attr_to_main,
                                                self.ID_BY_NAME_COL, wb)
                self.update_md5(md5, value)

            # related object data
            for idx, attr in enumerate(related_cs.attrs, idx + 1):
                value = cell(idx)
                value = self.update_import_data(
                    related_data, attr, value, related_attr_to_col,
                    related_attr_to_main, related_cs.ID_BY_NAME_COL, wb)
                self.update_md5(md5, value)

            # main item: tail data
            for idx, attr in enumerate(attrs, idx + 1):
                if is_new and attr == 'id':
                    continue
                value = cell(idx)
                value = self.update_import_data(data, attr, value, attr_to_col,
                                                attr_to_main,
                                                self.ID_BY_NAME_COL, wb)
                self.update_md5(md5, value)

            md5 = md5.hexdigest()
            if is_new or orig_md5 != md5:
                for attr in single_value_attrs:
                    if attr in related_data:
                        related_data[attr] = related_data[attr][attr]

                model_name = data[self.ID_BY_NAME_COL].get(self.ID_BY_NAME_COL)
                if not model_name:
                    continue

                model_data = name_map.get(model_name)
                if not model_data:
                    continue

                model_data.setdefault(related_cs.related_attr,
                                      []).append(related_data)

    def _get_modified_data(self, obj, data, attr_to_col, is_new, is_update):
        """
        After data is read and processed from XLS cells, compare with existing
        object to get only changed values.
        """
        modified = {}
        for attr, value in data.items():
            col = attr_to_col[attr]
            # import_val is current obj value formated for import, so that we can
            # compare with cell value
            import_val, current = col.data_for_import(obj)

            if not self.is_importable(obj, col, current):
                # for now we can't handle this type of values
                continue

            has_data_for_new = (is_new and (attr in data) and value)
            has_data_update = (
                not is_new and (value != import_val) and
                (value or import_val))  # last test ensures (None, u'', 0,
            #False, ...) are treated as equal
            #
            # FIXME: exception for bool columns to
            # check None != False?

            if is_update and not value:
                # update mode: we don't expect full records, and we don't erase
                # values with empty ones
                has_data_update = False

            if has_data_for_new or has_data_update:
                update = None
                try:
                    update = self._import_value(obj, col, current, value)
                except ExcelImportError, e:
                    update = col.UpdateCls(attr, current, value, None)
                    update.error = True
                    update.error_msg = e.message
                    update.update = e.imported_value
                except ValueError, e:
                    update = col.UpdateCls(attr, current, value, None)
                    update.error = True
                    update.error_msg = e.message

                if col.required:
                    if update is None:
                        # 'if update' may evaluate to False, beware to test existence with
                        # 'is None'
                        update = col.UpdateCls(attr, current, value, None)

                    if update.value is None:
                        if is_update:
                            # update mode: we don't expect full records, and we don't erase
                            # values with empty ones
                            continue
                        modified['__required_missing__'] = True
                        if not update.error:
                            # don't mask import error with 'required' message
                            update.error = True
                            update.error_msg = _(u'This field is required')

                if update:
                    modified[attr] = update

        return modified

    def save_data(self, data):
        """
        @param data: list of (id(int), attr_sig, modified(dict))
        """
        session = db.session()
        changed_items = 0
        skipped_items = 0
        created_items = 0
        error_happened = False
        ids = [update.id for update in data if update.id is not None]
        q = self.model_cls.query

        if ids:
            q.filter(self.model_cls.id.in_(ids)).all()

        attr_to_column = self.attr_to_column

        for item_update in data:
            is_new = item_update.id is None
            item = q.get(item_update.id) if not is_new else self.model_cls()
            signed_attrs = '{}.{}'.format(u';'.join(item_update.attrs),
                                          item_update.sig)
            if not self.signer.validate(signed_attrs):
                item_id = str(item_update.id) if not is_new else 'new'
                logger.debug('Validation failed, skipping item "%s"'
                             '\n'
                             'item_attrs="%s"'
                             '\n'
                             'attr_sig="%s"', item_id, repr(item_update.attrs),
                             repr(item_update.sig))
                skipped_items += 1
                continue

            try:
                with session.begin_nested():
                    for attr, value in item_update.data.items():
                        if attr not in item_update.attrs:
                            continue
                        col = attr_to_column[attr]
                        if hasattr(col, 'type_') and col.type_ is not None:
                            value = col.type_(value)

                        value = col.deserialize(value)
                        import_val, current = col.data_for_import(item)
                        update = self._import_value(item, col, current, value)
                        # at this stage we don't expect anymore import errors
                        assert not update.error

                        if col.required and not update.value:
                            # FIXME: missing param
                            raise ExcelImportError(u'missing required')

                        if update:
                            self._set_obj_value(item, attr, update)

                    db.session.add(item)

                    # import "many relateds" values
                    many_relateds = item_update.data.get('__many_related__', {})
                    many_relateds_map = {cs.related_attr: cs
                                         for cs in self.MANY_RELATED_CS}

                    for rel_attr, updates in many_relateds.items():
                        cs = many_relateds_map.get(rel_attr)
                        if not cs:
                            logger.error(
                                'Many relateds: columns set for "%s" not found',
                                rel_attr,
                                extra={'stack': True,})
                            continue

                        manager = cs.create_manager()
                        prop = sa.inspect(self.model_cls).attrs[cs.related_attr]
                        # FIXME: we assume relation is made with 1 attribute
                        prop = list(prop._reverse_property)[0]
                        prop_key = prop.key
                        del prop
                        rel_attr_to_col = self.get_attr_to_column(
                            cs,
                            map_related_attr=True)

                        for update in updates:
                            obj = cs.model_cls()
                            setattr(obj, prop_key, item)

                            for attr, value in update.data.items():
                                if attr not in update.attrs:
                                    continue
                                col = rel_attr_to_col[attr]
                                if hasattr(col,
                                           'type_') and col.type_ is not None:
                                    value = col.type_(value)

                                value = col.deserialize(value)
                                import_val, current = col.data_for_import(obj)
                                imported = manager._import_value(obj, col,
                                                                 current, value)
                                # at this stage we don't expect anymore import errors
                                assert not imported.error

                                if col.required and not imported.value:
                                    # FIXME: missing param
                                    raise ExcelImportError(u'missing required')

                                if imported:
                                    manager._set_obj_value(obj, attr, imported)

                            db.session.add(obj)

            except Exception as e:
                if isinstance(e, sa.exc.StatementError):
                    logger.error('Import error: %s%s\n%s',
                                 e.message,
                                 e.statement,
                                 pprint.pformat(e.params),
                                 exc_info=True)
                else:
                    logger.error('Import error: %s',
                                 unicode(e).encode('utf-8'),
                                 exc_info=True)
                    raise
                error_happened = True
                skipped_items += 1
                # skip this item
                continue
            else:
                if is_new:
                    created_items += 1
                else:
                    changed_items += 1

        if error_happened:
            #FIXME: include data for showing failed signatures, error during attr
            #conversion, etc
            logger.error('Excel import error', extra={'stack': True,})

        db.session.commit()
        return dict(changed_items=changed_items,
                    created_items=created_items,
                    skipped_items=skipped_items,
                    error_happened=error_happened)

    def update_md5(self, md5, value):
        """ Compute consistent md5 for exported and imported value.

        For example if model value is None, exported value may be u''.
        All values are also converted to utf-8 encoded string before hashing.
        """
        if value is None or isinstance(value, list):
            value = u''

        value = unicode(value).encode('utf-8')
        md5.update(value)

    def extract_signature(self, signed, signer=None):
        """ Return the signature part of a signed string. """
        if signer is None:
            signer = self.signer
        signature = signer.sep.join(signed.rsplit(signer.sep, 2)[-2:])
        logger.debug('extract_signature: "%s" from "%s"', repr(signature),
                     signed)
        return signature

    def columns_for(self, field, form):
        if field.name in self.SKIP_COLS:
            return None

        custom_columns = 'columns_set_{}'.format(field.name)
        if hasattr(self, custom_columns):
            return getattr(self, custom_columns)(field, form)

        if isinstance(field, ModelFieldList):
            return None

        if hasattr(self.model_cls, field.name):
            attr = field.name
            label = field.label.text
            db_col = self.db_columns.get(field.name)
            column_cls, type_ = self.column_type(attr, db_col)
            required = u'required' in field.flags
            return column_cls(attr, label, type_, required)

        return None

    def column_type(self, attr, db_col):
        column_cls = None
        type_ = None
        custom_type = 'column_type_{}'.format(attr)
        if hasattr(self, custom_type):
            column_cls, type_ = getattr(self, custom_type)(attr, db_col)

        if column_cls is None:
            column_cls = Column

        if db_col is None:
            # relationship
            mapper = self.mapper
            prop = mapper.relationships.get(attr)

            if prop is not None and prop.direction is sa.orm.interfaces.MANYTOONE:
                target_class = prop.mapper.class_

                if issubclass(target_class, BaseVocabulary):
                    column_cls = VocabularyColumn
                    type_ = target_class.query.active().by_label
                elif issubclass(target_class, PostalAddress):
                    column_cls = PostalAddressColumn
                    type_ = None

        elif type_ is None and db_col is not None:
            try:
                type_ = db_col.type.python_type
                if type_ is bool:
                    type_ = lambda v: v in (True, 'True', '1', 1)
                elif type_ in (datetime.date, datetime.datetime):
                    # native excel type, no need to set cast function
                    column_cls = (DateColumn if type_ is datetime.date else
                                  DateTimeColumn)
                    type_ = None
            except NotImplementedError:
                pass

        return column_cls, type_

    def is_importable(self, item, column, current):
        if not column.importable:
            return False

        attr_name = (column.related_attr if isinstance(column, RelatedColumnSet)
                     else column.attr)

        custom_import = 'import_{}'.format(attr_name)
        if hasattr(self, custom_import):
            return True

        mapper = sa.inspect(item.__class__)
        relationship = mapper.relationships.get(attr_name)
        if relationship is not None:
            target_class = relationship.mapper.class_

            if (relationship.direction is sa.orm.interfaces.MANYTOONE and
                    issubclass(target_class, BaseVocabulary)):
                return True

            return False  # we cannot handle relationships in a generic way

        if isinstance(current, (db.Model, list)):
            return False

        try:
            col = getattr(self.db_columns, attr_name)
        except AttributeError:
            pass
        else:
            if isinstance(col.type, JSONType):
                return False

        return True

    def _import_value(self, obj, column, current, data):
        """ Return a ::class `Update` instance.

        Lookup for method import_{attr}(obj, value). Expected returned value
        is (value_key, value). value_key should be data convertible to
        string (for web serialization) and should give the same value if
        _import_value is called with it as parameter 'data'

        If there's no method defined, value is returned as is.

        May raise ExcelImportError or ValueError with an explicit message

        For example given a string, update value could be en Entity
        """
        is_related = False
        attr_name = None

        if isinstance(data, Invalid):
            raise ExcelImportError(
                _(u'Invalid cell format, value: {value}').format(
                    value=data.value),
                imported_value=data)

        if isinstance(column, RelatedColumnSet):
            is_related = True
            attr_name = column.related_attr
        else:
            attr_name = column.attr

        custom_import = 'import_{}'.format(attr_name)
        if hasattr(self, custom_import):
            data, value = getattr(self, custom_import)(obj, data)
            return column.UpdateCls(attr_name, current, data, value)

        value = data

        if is_related:
            raise ExcelImportError(
                _(u'Cannot import related entity for attribute "{}" '
                  'without a custom importer method').format(attr_name),
                imported_value=value)

        try:
            col = getattr(self.db_columns, attr_name)
        except AttributeError:
            pass
        else:
            if 'choices' in col.info:
                # unicode(v) will make translatable values... translated
                rev_choices = {unicode(v): k
                               for k, v in col.info.get('choices').items()}
                if value not in rev_choices:
                    valid = u', '.join((u'"{}"'.format(v)
                                        for v in rev_choices.keys() if v))
                    raise ExcelImportError(
                        _(u'"{value}" is invalid. Valid choices are: {valid}'
                          ).format(value=value, valid=valid),
                        imported_value=value,
                    )
                value = rev_choices.get(value, value)

        return column.UpdateCls(attr_name, current, data, value)

    def _set_obj_value(self, obj, attr, update):
        """
        During save: set value in `update` on obj.

        This method is meant to be overriden for special attributes
        """
        setattr(obj, attr, update.value)

    def get_cell_value(self, sheet, row, col):
        """
        convert a cell value to useful python value returns a tuple of (value,
        excel cell type)
        """
        cell = sheet.cell(row=row, column=col)
        value = cell.value

        if isinstance(value, STRING_TYPES):
            value = cell.value.strip().replace(u'\r\n', u'\n')
            if u'\222' in value or u'\225' in value:
                # handle weird copy paste that may happen in excel: value is not actual
                # unicode but a cp1252 one that must be converted to unicode
                value = value.encode('raw_unicode_escape').decode('cp1252')
            elif value == '-':
                value = u""

        return value

    def update_import_data(self, data, attr, value, attr_map, attr_to_main,
                           ID_BY_NAME_COL, UNIQUE_ID_COLS, wb):
        """ Update data dict of imported value.

        (value, cell_types) comes from get_cell_value.

        @param wb: current workbook
        """
        column = attr_map[attr]

        if (value is not None and column.expected_cell_types is not None):
            if type(value) not in column.expected_cell_types:
                try:
                    value = column.adapt_from_cell(value, wb)
                except:
                    value = Invalid(value)

        if (not isinstance(value, Invalid) and value is not None and
                column.type_ is not None):
            try:
                value = column.type_(value)
            except:
                value = Invalid(value)

        main_col = attr_to_main[attr]
        main_col_name = (main_col.related_attr
                         if isinstance(main_col, RelatedColumnSet) else
                         column.attr)
        data[main_col_name][attr] = value

        if attr == ID_BY_NAME_COL and not isinstance(value, Invalid):
            data['__metadata__']['unique_name'] = value

        if attr in UNIQUE_ID_COLS:
            data['__metadata__']['unique_id_cols'][attr] = value

        return value



# coding=utf-8
"""
Export / import from excel files
"""
from __future__ import absolute_import

from .manager import ExcelManager
from .views import ExcelModuleComponent
from .columns import RelatedColumnSet, ManyRelatedColumnSet

__all__ = ("ExcelManager", "ExcelModuleComponent", "RelatedColumnSet",
           "ManyRelatedColumnSet")



# coding=utf-8
from __future__ import absolute_import

#noinspection PyUnresolvedReferences
from pkgutil import extend_path

#noinspection PyUnboundLocalVariable
__path__ = extend_path(__path__, __name__)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import jinja2
import pkg_resources

from abilian.core.signals import register_js_api
from abilian.core.util import fqcn
from abilian.web import url_for

from . import jinja_filters
from .excel.views import bp as excel_bp

STATIC_DIR = pkg_resources.resource_filename(__name__, 'static')
JS = ('js/async_export.js',)


class AbilianCRM(object):
    """
  Base extension required by abilian.crm.apps
  """

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if FQCN in app.extensions:
            return

        app.extensions[FQCN] = self
        app.register_blueprint(excel_bp, url_prefix='/crm/excel')

        # register i18n
        app.extensions['babel'].add_translations('abilian.crm')

        jinja_filters.init_filters(app)
        app.register_jinja_loaders(jinja2.PackageLoader(__name__, 'templates'))

        # crm static assets
        app.add_static_url('abilian/crm',
                           STATIC_DIR,
                           endpoint='abilian_crm_static')
        app.extensions['webassets'].append_path(
            STATIC_DIR, app.static_url_path + '/abilian/crm')

        app.register_asset('js', *JS)
        register_js_api.connect(self.register_js_api)

    def register_js_api(self, sender):
        app = sender
        js_api = app.js_api.setdefault('crm', {})
        js_api = js_api.setdefault('excel', {})
        js_api['taskStatusUrl'] = url_for('crm_excel.task_status')


FQCN = fqcn(AbilianCRM)
crm = AbilianCRM()



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

import abilian.web.forms.fields as awbff
import abilian.web.forms.widgets as aw_widgets

from ...forms import PostalAddressField, PostalAddressForm
from ...models import PostalAddress
from .base import Field, FormField
from .registry import form_field, model_field


@model_field
class _PostalAddressField(Field):
    __fieldname__ = 'PostalAddress'
    sa_type = sa.Integer
    default_ff_type = 'PostalAddressFormField'
    allow_multiple = True

    def get_model_attributes(self, *args, **kwargs):
        # column declared_attr
        col_name = self.name + '_id'

        if self.multiple:
            yield self.gen_m2m(*args, **kwargs)
            raise StopIteration

        def gen_column(cls):
            fk_kw = dict(
                name=u'{}_{}_fkey'.format(cls.__name__.lower(), col_name),
                use_alter=True,
                ondelete='SET NULL',)
            target_col = str(PostalAddress.id.parent.c.id)
            fk = sa.ForeignKey(target_col, **fk_kw)
            return sa.Column(col_name, sa.types.Integer(), fk)

        gen_column.func_name = col_name
        yield col_name, sa.ext.declarative.declared_attr(gen_column)

        # relationship declared_attr
        def gen_relationship(cls):
            kw = dict(uselist=False)
            local = cls.__name__ + '.' + col_name
            remote = str(PostalAddress.id)
            kw['primaryjoin'] = '{} == {}'.format(local, remote)
            return sa.orm.relationship(PostalAddress, **kw)

        gen_relationship.func_name = self.name
        yield self.name, sa.ext.declarative.declared_attr(gen_relationship)

    def gen_m2m(self, *args, **kwargs):
        model_name = self.model

        def gen_relationship(cls):
            src_name = cls.__tablename__
            local_src_col = model_name.lower() + '_id'
            local_target_col = 'postaladdress_id'
            tbl_name = src_name + '_' + self.name
            secondary_table = sa.Table(
                tbl_name,
                cls.metadata,
                sa.Column(local_src_col, sa.ForeignKey(cls.id)),
                sa.Column(local_target_col, sa.ForeignKey(PostalAddress.id)),
                sa.schema.UniqueConstraint(local_src_col, local_target_col),)

            rel_kw = dict(secondary=secondary_table)
            return sa.orm.relationship(PostalAddress, **rel_kw)

        gen_relationship.func_name = self.name
        return self.name, sa.ext.declarative.declared_attr(gen_relationship)


@form_field
class PostalAddressFormField(FormField):
    ff_type = PostalAddressField

    def get_type(self, *args, **kwargs):
        return (awbff.ModelFieldList if self.multiple else self.ff_type)

    def get_extra_args(self, *args, **kwargs):
        extra_args = super(PostalAddressFormField, self)\
                        .get_extra_args(*args, **kwargs)
        if self.multiple:
            extra_args['unbound_field'] = awbff.ModelFormField(
                PostalAddressForm)
            extra_args['min_entries'] = 1
            extra_args['population_strategy'] = 'update'
            extra_args['widget'] = aw_widgets.TabularFieldListWidget(
                template='widgets/model_fieldlist.html',)

        return extra_args



# coding=utf-8
"""
"""
from __future__ import absolute_import

import wtforms.fields
import wtforms.widgets as wtf_widgets

import abilian.web.forms.fields as awbff
import abilian.web.forms.widgets as aw_widgets
from abilian.web.forms.filters import strip

from .base import FormField
from .registry import form_field


@form_field
class TextField(FormField):

    def get_filters(self, *args, **kwargs):
        return (strip,)


@form_field
class BooleanField(FormField):
    ff_type = wtforms.fields.BooleanField

    def get_validators(self, *args, **kwargs):
        # default implementation may add 'required'
        return ()

    def setup_widgets(self, extra_args):
        super(BooleanField, self).setup_widgets(extra_args)

        if 'widget' not in extra_args:
            kwargs = {}
            options = self.data.get('widget_args', {})
            kwargs['on_off_mode'] = bool(options)
            if not isinstance(options, dict):
                options = {}

            kwargs['on_off_options'] = options
            extra_args['widget'] = aw_widgets.BooleanWidget(**kwargs)


@form_field
class DateField(FormField):
    ff_type = awbff.DateField


@form_field
class DateTimeField(FormField):
    ff_type = awbff.DateTimeField


@form_field
class DecimalField(FormField):
    ff_type = wtforms.fields.DecimalField


@form_field
class IntegerField(FormField):
    ff_type = wtforms.fields.IntegerField

    def __init__(self, model, data, *args, **kwargs):
        super(IntegerField, self).__init__(model, data, *args, **kwargs)
        if 'widget' not in data:
            data['widget'] = wtf_widgets.html5.NumberInput()


@form_field
class EmailField(TextField):

    def __init__(self, model, data, *args, **kwargs):
        super(EmailField, self).__init__(model, data, *args, **kwargs)
        if 'widget' not in data:
            data['widget'] = aw_widgets.EmailWidget()
        if 'view_widget' not in data:
            widget = data['widget']
            data['view_widget'] = (widget
                                   if isinstance(widget, aw_widgets.EmailWidget)
                                   else aw_widgets.EmailWidget())


@form_field
class URLField(TextField):

    def __init__(self, model, data, *args, **kwargs):
        super(URLField, self).__init__(model, data, *args, **kwargs)
        if 'widget' not in data:
            data['widget'] = wtf_widgets.html5.URLInput()
        if 'view_widget' not in data:
            data['view_widget'] = aw_widgets.URLWidget()


@form_field
class FileField(FormField):
    ff_type = awbff.FileField


@form_field
class ImageField(FileField):

    def __init__(self, model, data, *args, **kwargs):
        if 'widget' not in data:
            data['widget'] = aw_widgets.ImageInput()
        super(ImageField, self).__init__(model, data, *args, **kwargs)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from operator import attrgetter

from markupsafe import Markup
from wtforms.fields import FieldList

from abilian.core.entities import Entity

_NULL_MARK = object()


class Invalid(object):
    """
  Mark invalid values
  """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return u'{module}.{cls}(value={value})'.format(
            module=self.__class__.__module__,
            cls=self.__class__.__name__,
            value=repr(self.value),).encode('utf-8')

    def __unicode__(self):
        return u'Invalid: {}'.format(repr(self.value))

    def __str__(self):
        return unicode(self).encode('utf-8')


class Update(object):
    """
  Holds information about a value update.  Used in import_data()
  """
    error = False
    error_msg = None

    def __init__(self, attr, current, update, value=_NULL_MARK):
        """
    @param current: current value
    @param update: updated value as received in XLS file (Entity key for
    example)
    @param value: value to be set on object (Entity instance for example)
    """
        self.attr = attr
        self.current = current
        self.update = update
        self.value = value if value is not _NULL_MARK else update

    def __nonzero__(self):
        return (self.current != self.value) or self.error

    def _render(self, manager, value):
        form = manager.form
        field = getattr(form, self.attr)
        if isinstance(field, FieldList):
            for v in value:
                field.append_entry(v)
        else:
            field.data = field.object_data = value
        return Markup(field.render_view())

    def render_current(self, manager):
        return self._render(manager, self.current)

    def render_value(self, manager):
        return self._render(manager, self.value)


class Column(object):
    """
  a single column
  """
    # a column may be declared unconditionnaly not importable
    importable = True
    expected_cell_types = None
    adapt_cell_types = ()
    UpdateCls = Update

    def __init__(self, attr, label, type_, required=False, col_attr=None):
        """ type_: callable to convert an imported value to model one
    """
        self.attr = attr
        self.col_attr = col_attr if col_attr is not None else attr
        self.label = label
        self.type_ = type_
        self.required = required

    def __repr__(self):
        return (
            '{module}.{cls}(attr={attr!r}, label={label!r}, type_={type_!r}, '
            'required={required!r}) at 0x{id:x}'
            ''.format(module=self.__class__.__module__,
                      cls=self.__class__.__name__,
                      attr=self.attr,
                      label=self.label,
                      type_=self.type_,
                      required=self.required,
                      id=id(self)))

    def __iter__(self):
        yield self

    iter_flatened = __iter__

    @property
    def attrs(self):
        yield self.col_attr

    @property
    def labels(self):
        yield self.label

    @property
    def colspan(self):
        return 1

    def data(self, item):
        if item is None:
            yield None, None
            raise StopIteration

        def to_str(item):
            if isinstance(item, str):
                return item.decode('utf-8')
            elif isinstance(item, Entity):
                return item.name
            else:
                return unicode(item)

        value = attrgetter(self.attr)(item)
        import_value = item.display_value(self.attr, value=value)
        if isinstance(import_value, Entity):
            import_value = import_value.name
        elif isinstance(import_value, list):
            import_value = u"; ".join((to_str(i) for i in import_value))

        if isinstance(import_value, str):
            import_value = import_value.decode('utf-8')
        if isinstance(import_value, unicode):
            import_value = import_value.strip().replace(u'\r\n', u'\n')

        yield import_value, value

    def data_for_import(self, item):
        # data() returns an iterator of "length" 1, data_for_import must return a
        # value directly
        for data in self.data(item):
            return data

    def adapt_from_cell(self, value, cell_type, workbook):
        if cell_type not in self.adapt_cell_types:
            raise ValueError('Cannot adapt {}'.format(cell_type))
        return self._adapt_from_cell(value, cell_type, workbook)

    def _adapt_from_cell(self, value, cell_type, workbook):
        raise NotImplementedError

    # def serialize(self, value):
    #   return value

    def deserialize(self, value):
        if self.type_ is not None:
            value = self.type_(value)
        return value


class ColumnSet(object):
    """
  A set of columns to be added to current export / import
  """
    # a columnset may be declared unconditionnaly not importable
    importable = True
    expected_cell_types = None
    UpdateCls = Update

    def __init__(self, *columns):
        self.columns = []
        for c in columns:
            if isinstance(c, (tuple, list)):
                c = Column(*c)

            if not isinstance(c, (Column, ColumnSet)) and callable(c):
                c = c()

            assert isinstance(c, (Column, ColumnSet))
            self.columns.append(c)

    def __iter__(self):
        return iter(self.columns)

    def iter_flatened(self):
        for sub_columns in self:
            for c in sub_columns.iter_flatened():
                yield c

    @property
    def colspan(self):
        return sum(c.colspan for c in self.columns)

    @property
    def attrs(self):
        for c in self.columns:
            for attr in c.attrs:
                yield attr

    @property
    def labels(self):
        for c in self.columns:
            for label in c.labels:
                yield label

    def data(self, item):
        for c in self.columns:
            for result in c.data(item):
                yield result

    def data_for_import(self, item):
        """ Like ::meth `data`, but for import. The difference is that it must
    return a tuple like Column.data. In the generic context of a ColumnSet it
    has no sense, but a RelatedColumnSet may return the related entity.
    """
        return _NULL_MARK, _NULL_MARK

    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value



# coding=utf-8
"""
"""
from __future__ import absolute_import

from abilian.web.action import FAIcon
from abilian.web.forms.widgets import TextInput

from .jinja_filters import format_phonenumber


class PhoneNumberWidget(TextInput):
    pre_icon = FAIcon('phone')

    def render_view(self, field, **kwargs):
        data = field.data
        if not data:
            return u''

        return format_phonenumber(data, international=True)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import imp
import logging
import sys
from importlib import import_module

from pathlib import Path

from .codegen import CodeGenerator

logger = logging.getLogger(__name__)


def generate_module(fullname, **kw):
    """
    :param module: existing module

    :param name: module name for generated classes

    :param directory: directory where resides *.yml files.
    By default, directory `name` in `module` directory.
    """
    module = imp.new_module(fullname)

    parent_module, name = fullname.rsplit('.', 1)
    parent_module = import_module(parent_module)
    directory = Path(parent_module.__file__).parent
    directory = directory / name

    assert directory.exists() and directory.is_dir()
    module.__path__ = [str(directory)]

    for yml in directory.glob(u'*.yml'):
        logger.info('Loading: %s', yml)
        with yml.open('rt', encoding='utf-8') as f:
            gen = CodeGenerator(yaml_file=f, **kw)

        gen.init_vocabularies(module)
        gen.gen_model(module)
        gen.gen_form(module)

    return module


class GeneratedModelsFinder(object):
    """
    Module finder for generated models
    """

    def __init__(self):
        self.managed_modules = {}
        sys.meta_path.append(self)

    def manage_module(self, package, name, **kw):
        """
        Install module loader for 'package.name'
        """
        fullname = package + '.' + name
        self.managed_modules[fullname] = kw

    def find_module(self, fullname, path=None):
        """
        PEP 302 finder
        """
        if fullname in self.managed_modules:
            return self
        return None

    def load_module(self, fullname):
        """
        PEP 302 loader
        """
        if fullname not in self.managed_modules:
            raise ImportError

        module = generate_module(fullname, **self.managed_modules[fullname])
        module.__loader__ = self
        sys.modules[fullname] = module
        return module


_finder = GeneratedModelsFinder()
install_loader = _finder.manage_module



# coding=utf-8
"""
"""
from __future__ import absolute_import

import wtforms.fields
from flask_wtf.file import FileField
from flask_wtf.html5 import IntegerField, TelField
from wtforms.ext.sqlalchemy.fields import QuerySelectField

import abilian.web.forms.fields as awbff
from abilian.i18n import country_choices
from abilian.web.forms import widgets as abilian_widgets
from abilian.web.forms.filters import lowercase, strip, uppercase
from abilian.web.forms.validators import email, optional, required, url

__all__ = ('FORM_FIELDS', 'FORM_FILTERS', 'LIST_GENERATORS', 'VALIDATORS',
           'WIDGETS')

# Postgres crops identifiers names silently. This can be a problem with alembic
# autogenerated revsions since intended column name will differ from actual name
# in database
#
MAX_IDENTIFIER_LENGTH = 63

FORM_FIELDS = (wtforms.fields.SelectField,
               wtforms.fields.BooleanField,
               wtforms.fields.TextField,
               wtforms.fields.TextAreaField,
               wtforms.fields.FloatField,
               wtforms.fields.DecimalField,
               wtforms.fields.IntegerField,
               QuerySelectField,
               FileField,
               TelField,
               IntegerField,
               awbff.Select2Field,
               awbff.Select2MultipleField,
               awbff.DateField,)
FORM_FIELDS = {t.__name__: t for t in FORM_FIELDS}

VALIDATORS = {
    'email': email,
    'url': url,
    'required': required,
    'optional': optional,
}

FORM_FILTERS = {'strip': strip, 'uppercase': uppercase, 'lowercase': lowercase,}

WIDGETS = (abilian_widgets.BooleanWidget,
           abilian_widgets.URLWidget,
           abilian_widgets.EmailWidget,
           abilian_widgets.MoneyWidget,
           abilian_widgets.HoursWidget,)
WIDGETS = {t.__name__: t for t in WIDGETS}

LIST_GENERATORS = {'country': country_choices}


def update(module):
    """
    Install new column types, form fields, widgets, validators found in `module`.
    """
    for attr in __all__:
        values = getattr(module, attr, None)
        if values is None:
            continue

        definitions = globals()[attr]
        definitions.update(values)



# coding=utf-8
"""
Celery task for async export
"""
from __future__ import absolute_import

from cStringIO import StringIO
from itertools import ifilter
from time import gmtime, strftime
from urlparse import urlparse

from celery import shared_task
from flask import current_app, request
from flask_login import login_user
from werkzeug.utils import import_string

from abilian.core.models.subjects import User

from .util import XLSX_MIME

DEFAULT_EXPIRES = 1800  # generally user will not wait 1/2h. No need to process
# this long task


@shared_task(bind=True,
             track_started=True,
             ignore_result=False,
             expires=DEFAULT_EXPIRES,)
def export(self,
           app,
           module,
           from_url,
           user_id,
           component='excel',
           manager=None):
    """
    Async export xls task.

    :param app: `CRUDApp` name

    :param module: module id

    :param from_url: full url that started this task

    :param user_id: user's id for who this task is run
    """
    user = User.query.get(user_id)
    crud_app = current_app.extensions[app]
    module = crud_app.get_module(module)
    component = module.get_component(component)
    url = urlparse(from_url)
    rq_ctx = current_app.test_request_context(
        base_url=u'{url.scheme}://{url.netloc}/{url.path}'.format(url=url),
        path=url.path,
        query_string=url.query,)

    def progress_callback(exported=0, total=0, **kw):
        self.update_state(state='PROGRESS',
                          meta={'exported': exported,
                                'total': total})

    uploads = current_app.extensions['uploads']

    with rq_ctx:
        login_user(user, remember=False, force=False)
        objects = module.ordered_query(request)
        related_cs = None
        if manager is not None:
            manager = import_string(manager)
        else:
            manager = component.excel_manager

        manager = manager(module.managed_class, component.export_form,
                          component.EXCEL_EXPORT_RELATED)

        if 'related' in request.args:
            related = request.args['related']
            related_cs = ifilter(lambda cs: cs.related_attr == related,
                                 component.EXCEL_EXPORT_RELATED)
            try:
                related_cs = next(related_cs)
            except StopIteration:
                related_cs = None

        workbook = manager.export(objects,
                                  related_cs,
                                  progress_callback=progress_callback)
        fd = StringIO()
        workbook.save(fd)
        fd.seek(0)
        # save in uploads dir, return handle needed for download
        filename = u"{}-{}.xlsx".format(module.managed_class.__name__,
                                        strftime("%d:%m:%Y-%H:%M:%S", gmtime()))
        handle = uploads.add_file(user,
                                  fd,
                                  filename=filename,
                                  mimetype=XLSX_MIME)
        return dict(handle=handle)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging

import phonenumbers

from abilian.i18n import default_country

logger = logging.getLogger(__name__)


def format_phonenumber(n, international=True):
    """
  Format phone number for display.

  No formatting is applied if the number is not a valid phonenumber.

  :param international: always use international format, unless number is in
  national format OR country is the same as app's default country.
  """
    country = default_country() or u'FR'
    try:
        pn = phonenumbers.parse(n, country)
    except phonenumbers.NumberParseException:
        return n
    except:
        logger.exception('error while applying jinja filter "phonenumber" '
                         '- filter ignored')
        return n

    if not (phonenumbers.is_possible_number(pn) and
            phonenumbers.is_valid_number(pn)):
        return n

    fmt = phonenumbers.PhoneNumberFormat.INTERNATIONAL
    number_country = phonenumbers.region_code_for_country_code(pn.country_code)
    if not international and number_country == country:
        fmt = phonenumbers.PhoneNumberFormat.NATIONAL

    return phonenumbers.format_number(pn, fmt)


def init_filters(app):
    app.jinja_env.filters['phonenumber'] = format_phonenumber



# coding=utf-8
"""
"""
from __future__ import absolute_import

from mock import patch

from abilian.crm import jinja_filters


def test_format_phonenumber():
    fmt = jinja_filters.format_phonenumber

    with patch('abilian.crm.jinja_filters.default_country') as default_country:
        default_country.return_value = u'US'
        assert fmt(u'2025550166') == u'+1 202-555-0166'
        assert default_country.called
        assert fmt(u'202-555-0166') == u'+1 202-555-0166'
        assert fmt(u'2025550166', international=False) == u'(202) 555-0166'
        assert fmt(u'+33102030405') == u'+33 1 02 03 04 05'
        assert fmt(u'+33102030405', international=False) == u'+33 1 02 03 04 05'

        # "vanity" number
        # phonenumbers lib can parse it as a valid number. Maybe we could
        # display both, one for keeping the "vanity" form, the other for explicit
        # numbers.
        assert fmt(u'(800) Flowers') == u'+1 800-356-9377'
        assert fmt(u'1-800-Flowers', international=False) == u'(800) 356-9377'

        default_country.reset_mock()
        default_country.return_value = u'FR'
        assert fmt(u'+12025550166') == u'+1 202-555-0166'
        assert fmt(u'(00)12025550166') == u'+1 202-555-0166'
        assert fmt(u'+33102030405') == u'+33 1 02 03 04 05'
        assert fmt(u'+33102030405', international=False) == u'01 02 03 04 05'
        assert fmt(u'0102030405', international=False) == u'01 02 03 04 05'

        # we want to avoid NumberParseException
        with patch('abilian.crm.jinja_filters.logger.exception') as logger:
            assert fmt(
                u'garbage long unparsable text &"é(§è!çà)-)') == u'garbage long unparsable text &"é(§è!çà)-)'
            assert not logger.called



# coding=utf-8
"""
Set up listener to automatically set 'name' from various attributes.
"""
from __future__ import absolute_import

import string
from operator import attrgetter

import sqlalchemy as sa


def setup(cls, spec):
    """
    :param cls: an Entity based class
    :param spec: a valid python format string
    """
    parser = string.Formatter()
    attributes = {
        field: attrgetter(field)
        for text, field, format_spec, conversion in parser.parse(spec) if field
    }

    # event handler
    def auto_name(obj, new_val, old_val, initiator):
        """ Auto set name. """
        if new_val != old_val:
            key = initiator.key
            vals = {key: new_val}
            for attr, getter in attributes.iteritems():
                if attr != key:
                    val = getter(obj)
                    if val is None:
                        val = u''
                    vals[attr] = val

            obj.name = spec.format(**vals).strip()

    # install listener on attributes
    for attr in attributes.iterkeys():
        attr_impl = getattr(cls, attr)
        sa.event.listen(attr_impl, "set", auto_name,
                        propagate=True,  # set listener on subclasses
                        active_history=True,
                        retval=False)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from functools import partial


class Registrable(object):
    #: if not `None`, this is used as __fieldtype__()
    __fieldname__ = None

    @classmethod
    def __fieldtype__(cls):
        """
        :return: identifier name for this field type
        """
        return (cls.__fieldname__ if cls.__fieldname__ is not None else
                cls.__name__)


_SA_FIELD_REGISTRY = dict()
_FF_FIELD_REGISTRY = dict()


def _register(registry, cls):
    """
    class decorator for `Registrable` subclasses
    """
    assert issubclass(cls, Registrable)
    reg_attr = '_{}_registered'.format(cls.__name__)
    if getattr(cls, reg_attr, False):
        return cls
    name = cls.__fieldtype__()

    assert name not in registry, (
        '{!r} cannot be registered as {!r}: already used by {!r}'
        ''.format(cls, name, registry[name]))
    registry[name] = cls
    setattr(cls, reg_attr, True)
    return cls


model_field = partial(_register, _SA_FIELD_REGISTRY)
form_field = partial(_register, _FF_FIELD_REGISTRY)

get_field = _SA_FIELD_REGISTRY.get
get_formfield = _FF_FIELD_REGISTRY.get



# coding=utf-8
"""
"""
from __future__ import absolute_import

from openpyxl.cell import STRING_TYPES

from .base import Column

__all__ = ('VocabularyColumn',)


class VocabularyColumn(Column):
    """
  Columns for :class:`abilian.services.models.BaseVocabulary` items
  """
    expected_cell_types = STRING_TYPES

    def data(self, item):
        value = getattr(item, self.attr, None)
        import_value = unicode(value) if value is not None else u''
        yield import_value, value

    def deserialize(self, value):
        # type_ has already 'deserialized' value
        return value



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging
import re
from collections import OrderedDict

import sqlalchemy as sa
import yaml

from abilian.core.entities import Entity
from abilian.core.models import attachment, comment, tag
from abilian.core.util import slugify
from abilian.services.security import (CREATE, DELETE, READ, WRITE, Permission,
                                       Role)
from abilian.services.vocabularies import Vocabulary, get_vocabulary
from abilian.web.forms import Form, FormPermissions
from abilian.web.tags.extension import ns as tag_ns

from . import autoname
from .fieldtypes import get_field

logger = logging.getLogger(__name__)


def assert_ascii(s):
    try:
        s.encode('ascii')
    except UnicodeError:
        raise ValueError('{} is not an ASCII string'.format(repr(s)))


class CodeGenerator(object):

    def __init__(self, yaml_file=None, data=None, **options):
        self.vocabularies = {}
        self.options = options
        self._model_finalizers = []
        if data is not None:
            self.data = data
        else:
            self.load_file(yaml_file)
        self.prepare_data()

    def load_file(self, yaml_file):
        self.data = yaml.load(yaml_file)

    def add_model_finalizer(self, finalizer):
        self._model_finalizers.append(finalizer)

    def set_current_module(self, module):
        self.module = module

    def prepare_data(self):
        """Massage data before models creations."""
        for d in self.data['fields']:
            vocabulary = d.get('vocabulary')
            if vocabulary:
                name = u'Vocabulary_'
                group = vocabulary.get('group', u'').strip()
                if group:
                    group = slugify(group, u'_')
                    name += group + u'__'
                name += vocabulary['name']

                if name in self.vocabularies:
                    # already defined in another field, maybe on another model: share
                    # existing definition so that generated class is also accessible
                    d['vocabulary'] = self.vocabularies[name]
                else:
                    vocabulary['generated_name'] = name.encode('ascii')
                    self.vocabularies[name] = vocabulary
                continue

            # slugify list items
            from_list = d.get('from_list')
            if from_list is None:
                continue

            key_val_list = []
            seen = dict()
            for item in from_list:
                k = v = item
                if isinstance(item, (list, tuple)):
                    k, v = item
                    if isinstance(k, str):
                        k = k.decode('utf-8')
                    v = unicode(v)
                else:
                    if not isinstance(item, unicode):
                        item = item.decode('utf-8')
                    k = unicode(slugify(item, u'_'))

                k = re.sub(u'[^\w\s-]', '', k).strip().upper()
                k = re.sub(u'[-\s]+', '_', k)
                #  avoid duplicates: suffix by a number if needed
                current = k
                count = 0
                while k in seen and seen[k] != v:
                    count += 1
                    k = '{}_{}'.format(current, count)

                seen[k] = v
                key_val_list.append((k, v))

            d['from_list'] = key_val_list

    def init_vocabularies(self, module):
        self.set_current_module(module)
        for generated_name, definition in self.vocabularies.items():
            name = definition['name'].encode('ascii').strip()
            group = definition.get('group', u'').strip() or None
            label = definition['label'].strip()
            voc_cls = get_vocabulary(name, group=group)

            if voc_cls is None:
                voc_cls = Vocabulary(name=name, group=group, label=label)

            definition['cls'] = voc_cls

            if not hasattr(module, generated_name):
                setattr(module, generated_name, voc_cls)

    def gen_model(self, module):
        self.set_current_module(module)
        table_args = self.data.get('table_args', [])
        model_name = self.data['name']
        type_name = self.data.get('type_name', model_name + 'Base')
        type_base = self.data.get('type_base', Entity)
        try:
            type_base_attrs = sa.inspect(type_base).attrs
        except sa.exc.NoInspectionAvailable:
            type_base_attrs = frozenset()

        attributes = OrderedDict(self.data.get('attributes', {}))
        attributes['__module__'] = module.__name__
        attributes['__tablename__'] = self.data.get('tablename',
                                                    model_name.lower())

        # default permissions
        permissions = self.data.get('permissions')
        if permissions:
            default_permissions = {}
            for perm, roles in permissions.items():
                perm = perm.strip()
                if not perm:
                    raise TypeError('Found empty string for permission')
                perm = Permission(perm)
                roles = {Role(r.strip()) for r in roles if r.strip()}

                if not roles:
                    continue

                default_permissions[perm] = roles
        else:
            default_permissions = self.options.get('default_permissions', {})

        if default_permissions:
            for p in (WRITE, CREATE, DELETE):
                if p in default_permissions:
                    read_permissions = default_permissions.setdefault(READ,
                                                                      set())
                    read_permissions |= default_permissions[p]

            if WRITE in default_permissions:
                for p in (CREATE, DELETE):
                    if p not in default_permissions:
                        default_permissions[p] = set(default_permissions[WRITE])

            attributes['__default_permissions__'] = default_permissions

        # Fields
        for d in self.data['fields']:
            if 'ignore' in d:
                continue

            type_ = d['type']
            if type_ == 'pass':
                # explicit manual handling - only declared in form groups
                continue

            FieldCls = get_field(type_)
            if FieldCls is None:
                raise ValueError('Unknown type: {}'.format(repr(type_)))

            field = FieldCls(model=model_name, data=d, generator=self)

            if d['name'] in type_base_attrs:
                # existing field (i.e, Entity.name), don't override column else it will
                # be duplicated in joined table, and missing some setup found in
                # overriden column (like indexability)
                #
                # since field has been setup, it has created 'formfield' instance,
                # required for form generation
                continue

            for name, attr in field.get_model_attributes():
                assert name not in attributes
                attributes[name] = attr

            for arg in field.get_table_args():
                table_args.append(arg)

        for finalize in self._model_finalizers:
            finalize(attributes, table_args, module)

        if table_args:
            attributes['__table_args__'] = tuple(table_args)

        cls = type(type_name, (type_base,), attributes)
        self.data['cls'] = cls
        setattr(module, type_name, cls)

        # automatic name from other fields
        auto_name = unicode(self.data.get('auto_name') or u'').strip()

        if auto_name:
            autoname.setup(cls, auto_name)

        # commentable?
        if self.data.get('commentable', False):
            comment.register(cls)

        # attachments support ?
        if self.data.get('attachments', False):
            attachment.register(cls)

        # tagging support ?
        cls_tags_ns = self.data.get('tag')
        if cls_tags_ns:
            tag.register(cls)
            tag_ns(cls_tags_ns)(cls)

        return cls

    def gen_form(self, module):
        self.set_current_module(module)
        type_name = self.data['name'] + 'EditFormBase'
        type_bases = (Form,)
        attributes = OrderedDict()

        groups = {}
        group_names = []
        read_permissions = {}
        write_permissions = {}

        for d in self.data['fields']:
            if 'ignore' in d or 'hidden' in d:
                continue

            field_name = d['name']

            if 'ignore' not in d:
                group_name = d.get('group', 'default group')
                groups.setdefault(group_name, []).append(field_name)
                if group_name not in group_names:
                    group_names.append(group_name)

                perms = d.get('permissions', {})
                write = {Role(r.strip()) for r in perms.get('write', ())}
                read = write | {Role(r.strip()) for r in perms.get('read', ())}
                if write:
                    write_permissions[field_name] = write
                if read:
                    read_permissions[field_name] = read

            if d['type'] == 'pass':
                # explicit manual handling - only declared in form groups
                continue

            field = d['formfield']
            for name, attr in field.get_form_attributes():
                attributes[name] = attr

        attributes['_permissions'] = FormPermissions(
            fields_read=read_permissions,
            fields_write=write_permissions,)
        attributes['_groups'] = OrderedDict((name, groups[name])
                                            for name in group_names)
        attributes['__module__'] = module.__name__
        cls = type(type_name, type_bases, attributes)
        setattr(module, type_name, cls)
        return cls



# coding=utf-8
"""
"""
from __future__ import absolute_import

import hashlib

import sqlalchemy as sa

from abilian.core.models.blob import Blob
from abilian.core.sqlalchemy import JSON, JSONDict, JSONList

from ..definitions import MAX_IDENTIFIER_LENGTH
from .base import Field
from .registry import model_field


@model_field
class Integer(Field):
    """
    4 bytes: -2147483648 to +2147483647
    """
    sa_type = sa.types.Integer
    default_ff_type = 'IntegerField'


@model_field
class SmallInteger(Integer):
    """
    2 bytes: -32768 to +32767
    """
    sa_type = sa.types.SmallInteger


@model_field
class BigInteger(Integer):
    """
    8 bytes: -9223372036854775808 to 9223372036854775807
    """
    sa_type = sa.types.BigInteger


@model_field
class PositiveInteger(Integer):
    """
    :class:`Integer` restricted to positive value: 0 to +2147483647
    """

    def get_table_args(self, *args, **kwargs):
        col_name = self.name[:MAX_IDENTIFIER_LENGTH]
        name = 'check_{name}_positive'.format(name=col_name)
        if len(name) > MAX_IDENTIFIER_LENGTH:
            digest = hashlib.md5(self.name).digest()
            exceed = len(name) - MAX_IDENTIFIER_LENGTH
            name = self.name[:MAX_IDENTIFIER_LENGTH - exceed - 7]
            name = 'check_{name}_{digest}_positive'.format(name=name,
                                                           digest=digest[:6])

        yield sa.schema.CheckConstraint(
            sa.sql.text(col_name + ' >= 0'),
            name=name)


@model_field
class UnicodeText(Field):
    sa_type = sa.types.UnicodeText


@model_field
class LargeBinary(Field):
    sa_type = sa.types.LargeBinary


@model_field
class Date(Field):
    sa_type = sa.types.Date
    default_ff_type = 'DateField'


@model_field
class DateTime(Field):
    sa_type = sa.types.DateTime
    default_ff_type = 'DateTimeField'


@model_field
class Text(Field):
    sa_type = sa.types.Text


@model_field
class Float(Field):
    sa_type = sa.types.Float
    default_ff_type = 'DecimalField'


@model_field
class Boolean(Field):
    sa_type = sa.types.Boolean
    default_ff_type = 'BooleanField'


@model_field
class File(Field):
    sa_type = sa.types.Integer
    default_ff_type = 'FileField'
    allow_multiple = False

    def get_model_attributes(self, *args, **kwargs):
        col_name = self.name[:(MAX_IDENTIFIER_LENGTH - 3)] + '_id'
        extra_args = {'nullable': not self.required}
        extra_args['info'] = info = {}
        info['label'] = self.label

        attr = sa.schema.Column(col_name, self.sa_type(**self.sa_type_options),
                                sa.ForeignKey(Blob.id), **extra_args)

        yield col_name, attr

        relationship = sa.orm.relationship(
            Blob,
            primaryjoin='{tablename}.c.{local} == Blob.id'.format(
                tablename=self.model.lower(),
                local=col_name),)
        yield self.name, relationship


@model_field
class Image(File):
    default_ff_type = 'ImageField'


@model_field
class JSON(Field):
    sa_type = JSON


@model_field
class JSONListField(Field):
    __fieldname__ = 'JSONList'
    sa_type = staticmethod(JSONList)


@model_field
class JSONDict(Field):
    sa_type = staticmethod(JSONDict)


@model_field
class EmailAddress(UnicodeText):
    default_ff_type = 'EmailField'


@model_field
class URL(UnicodeText):
    default_ff_type = 'URLField'



# coding=utf-8
"""
Field for yearly data

Use to add attributes values by year; might be several in a row. If model has
other yearly attributes, they will be in the same row (from database POV).

In yml they are specified like this:

::
    - name: attr
      type: Yearly
      type_args:
        fields:
          - name: attr_1
            type: Integer
          - name: ...

"""
from __future__ import absolute_import

from functools import total_ordering
from operator import attrgetter

import sqlalchemy as sa
from sqlalchemy.orm.collections import collection
from wtforms.fields import FieldList, IntegerField
from wtforms.utils import unset_value as unset_value

import abilian.web.forms.fields as awbff
import abilian.web.forms.widgets as aw_widgets
from abilian.core.extensions import db
from abilian.core.models import SYSTEM
from abilian.i18n import _l
from abilian.web.forms import Form

from .base import FormField as FormFieldGeneratorBase
from .base import Field
from .registry import form_field, model_field

_MARK = object()


@total_ordering
class YearlyBase(db.Model):
    """
    Base model for yearly data collections
    """
    __abstract__ = True

    id = sa.Column(sa.Integer, primary_key=True, info=SYSTEM)
    year = sa.Column(sa.SmallInteger, nullable=False)

    def __eq__(self, other):
        raise NotImplementedError

    def __lt__(self, other):
        raise NotImplementedError


class YearlyCollection(sa.orm.collections.MappedCollection):
    """
    `sa.orm.relationships` collection class for yearly values.

    Years are used as keys.
    """

    def __init__(self, yearly_cls):
        super(YearlyCollection, self).__init__(YearlyCollection._keyfunc)
        self.yearly_cls = yearly_cls

    @collection.internally_instrumented
    def __setitem__(self, key, value, _sa_initiator=None):
        if not isinstance(key, int):
            raise ValueError('Key must be an integer, key={!r}'.format(key))
        value = self._value_fromdict(key, value)
        return super(YearlyCollection, self).__setitem__(key, value,
                                                         _sa_initiator)

    @collection.internally_instrumented
    def setdefault(self, key, default=None):
        """
        D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D.
        """
        if key not in self:
            self.__setitem__(key, default)

        # dont't return default as passed but value as actually set.
        # allow to pass 'default' as dict and get a YearlyBase class in return
        return self.__getitem__(key)

    @collection.converter
    def _convert(self, dictlike):
        for incoming_key, value in sa.orm.util.dictlike_iteritems(dictlike):
            yield self._value_fromdict(incoming_key, value)

    def _value_fromdict(self, key, value):
        if not isinstance(value, YearlyBase):
            if not isinstance(value, dict):
                raise ValueError(
                    'Value must a YearlyBase based class or a dict')

            value['year'] = key
            value = self.yearly_cls(**value)
        else:
            value_key = self.keyfunc(value)
            if key != value_key:
                raise TypeError(
                    "Found incompatible key {!r} for value {!r}; this collection's "
                    "keying function requires a key of {!r} for this value."
                    "".format(key, value, value_key))

        return value

    def keys(self):
        return sorted(super(YearlyCollection, self).keys())

    def iterkeys(self):
        for key in self.keys():
            yield key

    __iter__ = iterkeys

    def values(self):
        return sorted(
            super(YearlyCollection, self).values(),
            key=YearlyCollection._keyfunc)

    def itervalues(self):
        for v in self.values():
            yield v

    def items(self):
        return [(obj.year, obj) for obj in self.itervalues()]

    def iteritems(self):
        for item in self.items():
            yield item

    def latest(self, attr):
        """
        Return the most recent tuple `(year, value)` for given :attr:
        """
        getter = attrgetter(attr)
        for year, infos in reversed(self.items()):
            val = getter(infos)
            if val is not None:
                return (year, infos)

        return (None, None)

    @staticmethod
    def _keyfunc(obj):
        return obj.year


class YearlyAttrProxy(object):
    """
    Proxy model to allow get and update a particular attribute on yearly
    collections, as if it was directly a model collection.
    """
    __slots__ = ('yearly_data', 'attrs')

    def __init__(self, yearly_data, attrs):
        self.attrs = frozenset(attrs)
        self.yearly_data = yearly_data

    def clear(self):
        for attr in self.attrs:
            setattr(self.yearly_data, attr, None)

    def update(self, value):
        if isinstance(value, YearlyAttrProxy):
            for attr in self.attrs:
                setattr(self, attr, getattr(value, attr))
        elif isinstance(value, dict):
            for attr in self.attrs:
                v = value.get(attr)
                if v is not None:
                    setattr(self, attr, v)
        else:
            raise TypeError(
                'Except a YearlyAttrProxy or dict instance, got: {!r}'
                ''.format(type(value).__name__))

    def __getattr__(self, name):
        if name != 'attrs' and (name == 'year' or name in self.attrs):
            return self.yearly_data.__getattribute__(name)

        raise AttributeError

    def __setattr__(self, name, value):
        if name != 'attrs' and (name == 'year' or name in self.attrs):
            return self.yearly_data.__setattr__(name, value)

        return super(YearlyAttrProxy, self).__setattr__(name, value)

    def __nonzero__(self):
        return any(getattr(self.yearly_data, attr, None) is not None
                   for attr in self.attrs)

    def __repr__(self):
        return '<{} for {}: {!r}) at 0x{:x}>'.format(
            self.__class__.__name__, self.yearly_data.__class__.__name__,
            tuple(sorted(self.attrs)), id(self))

# proxy sa.inspect()
sa.inspection._inspects(YearlyAttrProxy)(
    lambda target: sa.inspect(target.yearly_data))


class YearlyCollectionProxy(dict):
    """
    Proxy the collection with :class:`YearlyAttrProxy` instances as values.

    New keys insert new :class:`YearlyAttrProxy` instances
    """

    def __init__(self, collection, attrs):
        self.__collection = collection
        self.__attrs = frozenset(attrs)

        for year, year_data in collection.iteritems():
            value = YearlyAttrProxy(year_data, self.__attrs)
            if value:
                self.__setitem__(year, value)

    def __getitem__(self, key):
        if key not in self.__collection:
            if key in self:
                super(YearlyCollectionProxy, self).__delitem__(key)
            raise KeyError

        year_data = self.__collection[key]
        if key not in self:
            super(YearlyCollectionProxy, self).__setitem__(
                key, YearlyAttrProxy(year_data, self.__attrs))

        return super(YearlyCollectionProxy, self).__getitem__(key)

    def __setitem__(self, key, value):
        if isinstance(value, YearlyAttrProxy):
            if value.attrs != self.__attrs:
                raise ValueError('Incompatible attribute proxys: {!r}, {!r}'
                                 ''.format(self, value))
            if key not in self and value:
                self[key]  # add key since value is nonzero
                return  # same proxy: no update needed

        elif isinstance(value, dict):
            if set(value) - self.__attrs:
                raise ValueError(
                    'Some keys are invalid for this collection proxy: {!r} (valid: {!r})'
                    ''.format(
                        sorted(set(value) - self.__attrs), sorted(
                            self.__attrs)))
            if key not in self.__collection:
                self.__collection[key] = {}

        proxy = self[key]
        proxy.update(value)

    def __delitem__(self, key):
        value = self.__getitem__(key, None)
        if value is not None:
            value.clear()
            del value
        super(YearlyCollectionProxy, self).__delitem__(key)

    def setdefault(self, key, default=None):
        'D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D'
        if key not in self.__collection:
            self.__setitem__(key, default)

        # dont't return default as passed but value as actually set.
        # allow to pass 'default' as dict and get a YearlyBase class in return
        return self.__getitem__(key)

    def keys(self):
        return sorted(super(YearlyCollectionProxy, self).keys())

    def iterkeys(self):
        for key in self.keys():
            yield key

    __iter__ = iterkeys

    def values(self):
        return sorted(
            super(YearlyCollectionProxy, self).values(),
            key=YearlyCollection._keyfunc)

    def itervalues(self):
        for v in self.values():
            yield v

    def items(self):
        return [(obj.year, obj)
                for obj in sorted(self.itervalues(),
                                  key=YearlyCollectionProxy._keyfunc)]

    def iteritems(self):
        for item in self.items():
            yield item


class YearlyAttribute(object):
    """
    An association proxy that allow multiple attributes grouping
    Descriptor for single attribute access
    """
    __slots__ = ('_attrs', '_collection_proxy')

    def __init__(self, attrs):
        self._attrs = frozenset(attrs)

    def __get__(self, instance, owner):
        if instance is None:
            # attribute accessed on class ('owner')
            return self

        return YearlyCollectionProxy(instance.__yearly_data__, self._attrs)

    def __set__(self, instance, value):
        yearly_data = instance.__yearly_data__
        for year, year_data in yearly_data.iteritems():
            if year not in value:
                for attr in self._attrs:
                    setattr(year_data, attr, None)

        for year, values in value.iteritems():
            year_data = yearly_data.setdefault(year, {})
            for attr in self._attrs:
                val = values.get(attr, _MARK)
                if val is not _MARK:
                    setattr(year_data, attr, val)

    def __delete__(self, instance):
        for year, year_data in instance.__yearly_data__.iteritems():
            for attr in self._attrs:
                setattr(year_data, attr, None)

    def __repr__(self):
        return '<{}{!r} at 0x{:x}>'.format(self.__class__.__name__, self._attrs,
                                           id(self))


@model_field
class Yearly(Field):

    default_ff_type = 'YearlyFormField'

    def __init__(self, model, data, generator, *args, **kwargs):
        super(Yearly, self).__init__(model, data, generator, *args, **kwargs)
        self.data['type_args']['name'] = self.name
        generator.add_model_finalizer(self.finalize)

        if 'yearly' not in generator.data:
            model_lower = model.lower()
            related_attr_id = '{}_id'.format(model_lower)
            tablename = '{}_yearly'.format(model_lower)
            type_name = '{}YearlyData'.format(model)
            fk_col = sa.Column(sa.Integer(),
                               sa.ForeignKey(self.model.lower() + '.id'))
            attributes = dict()
            attributes[related_attr_id] = fk_col
            attributes['_' + model_lower] = sa.orm.relationship(
                model,
                primaryjoin=
                u'{tablename}.c.{related_attr_id} == {remote}.c.id'.format(
                    tablename=tablename,
                    related_attr_id=related_attr_id,
                    remote=self.model.lower()))
            attributes['__auditable_entity__'] = ('_' + model_lower,
                                                  '__yearly_data__', ('year',))

            generator.data['yearly'] = dict(
                name=type_name,
                type_name=type_name,
                type_base=YearlyBase,
                tablename=tablename,
                related_attr_id=related_attr_id,
                related_attr=model_lower,
                cls=None,
                table_args=[sa.schema.UniqueConstraint(related_attr_id, 'year')
                           ],
                attributes=attributes,
                fields=[])

        self.yearly_data = generator.data['yearly']

    def get_model_attributes(self, *args, **kwargs):
        forbidden_attrs = {'id', 'year',}
        fields = self.data['type_args']['fields']
        if any(f['name'] in forbidden_attrs for f in fields):
            raise ValueError('{} are forbidden field names'.format(','.join(
                sorted(forbidden_attrs))))

        self.yearly_data['fields'].extend(fields)
        yield self.name, YearlyAttribute(f['name'] for f in fields)

    def finalize(self, attributes, table_args, module, *arg, **kwargs):
        """Create related model and relationships."""
        if self.yearly_data.get('cls') is None:
            self.create_related_model(module)

        yearly_cls = self.yearly_data['cls']
        primaryjoin = u'{local} == {remote}'.format(
            local=u'{__tablename__}.c.id'.format(**attributes),
            remote=u'{tablename}.c.{related_attr_id}'.format(**
                                                             self.yearly_data),)
        attributes['__yearly_data__'] = sa.orm.relationship(
            yearly_cls,
            primaryjoin=primaryjoin,
            collection_class=lambda: YearlyCollection(yearly_cls),
            cascade='all, delete-orphan',)

    def create_related_model(self, module):
        # cannot import at top of the file
        from ..codegen import CodeGenerator

        generator = CodeGenerator(data=self.yearly_data)
        generator.add_model_finalizer(self.yearly_model_finalizer)
        generator.init_vocabularies(module)
        rel_cls = generator.gen_model(module)
        self.yearly_data['cls'] = rel_cls
        return rel_cls

    def yearly_model_finalizer(self, attributes, *args, **kwargs):
        """Implements total_ordering methods."""
        rel_attr_id = self.yearly_data['related_attr_id']
        column_attributes = [attr
                             for attr, definition in attributes.iteritems()
                             if isinstance(definition, sa.Column)]

        def _eq(self, other):
            return (isinstance(other, self.__class__) and
                    all(getattr(self, attr) == getattr(other, attr)
                        for attr in column_attributes))

        def _lt(self, other):
            if getattr(self, rel_attr_id, -1) < getattr(other, rel_attr_id, -1):
                # dummy case, comparing objects not related to the same entity instance
                return True
            return self.year < other.year

        attributes['__eq__'] = _eq
        attributes['__lt__'] = _lt


class YearlyFieldList(awbff.ModelFieldList):

    def process(self, formdata, data=unset_value):
        if data is not unset_value and isinstance(data, dict):
            data = data.values()
        return super(YearlyFieldList, self).process(formdata, data)

    def _add_entry(self, formdata=None, data=unset_value, index=None):
        return FieldList._add_entry(self,
                                    formdata=formdata,
                                    data=data,
                                    index=index)

    def populate_obj(self, obj, name):
        entities = {}
        for entry in self.entries:
            data = entry.data
            year = data['year']
            entities[year] = data

        setattr(obj, name, entities)


@form_field
class YearlyFormField(FormFieldGeneratorBase):
    ff_type = YearlyFieldList

    def __init__(self, *args, **kwargs):
        super(YearlyFormField, self).__init__(*args, **kwargs)

    def get_extra_args(self, *args, **kwargs):
        from ..codegen import CodeGenerator

        generator = CodeGenerator(data=self.data['type_args'])
        year_field = IntegerField(label=_l(u'Year'))
        FormBase = generator.gen_form(self.generator.module)
        ModelField = type(self.name + 'Form', (FormBase,
                                               Form,), {'year': year_field,})

        extra_args = super(YearlyFormField, self).get_extra_args(*args,
                                                                 **kwargs)
        extra_args['unbound_field'] = awbff.FormField(ModelField, default=dict)
        extra_args['population_strategy'] = 'update'
        extra_args['min_entries'] = 1
        return extra_args

    def setup_widgets(self, extra_args):
        extra_args['widget'] = aw_widgets.TabularFieldListWidget()
        extra_args['view_widget'] = aw_widgets.ModelListWidget()



# coding=utf-8
"""
"""
from __future__ import absolute_import

from .registry import model_field, form_field, get_field, get_formfield

# register fields
from . import (simple, formfields, vocabulary, entity, yearly, postaladdress,
               phonenumber)

__all__ = ("model_field", "form_field", "get_field", "get_formfield", "simple",
           "formfields", "vocabulary", "entity", "yearly", "postaladdress",
           "phonenumber")



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

import abilian.web.forms.fields as awbff
import abilian.web.forms.widgets as aw_widgets
from abilian.web.action import Endpoint

from .base import Field, FormField
from .registry import form_field, model_field


@model_field
class EntityField(Field):
    __fieldname__ = 'Entity'
    default_ff_type = 'EntityFormField'

    def __init__(self, model, data, *args, **kwargs):
        super(EntityField, self).__init__(model, data, *args, **kwargs)
        self.target_cls = str(data.get('target', 'Entity').strip()) or 'Entity'

    def get_model_attributes(self, *args, **kwargs):
        target_col = self.target_cls.lower() + '.id'  # use tablename + '.id'
        col_name = self.name + '_id'
        type_args = self.data.get('type_args', {})

        res_iter = (self._m2m_relationship(target_col, col_name, type_args)
                    if self.multiple else
                    self._single_relationship(target_col, col_name, type_args))

        for result in res_iter:
            yield result

    def _single_relationship(self, target_col, col_name, type_args):
        # column
        def get_column_attr(func_name, col_name, target_cls_name, target_col):

            def gen_column(cls):
                # always use ALTER: when 2 entities reference each other it raises
                # CircularDependencyError:
                fk_kw = dict(
                    use_alter=True,
                    name='{}_{}_fkey'.format(cls.__name__.lower(), col_name),)

                fk_kw['ondelete'] = 'SET NULL'
                fk = sa.ForeignKey(target_col, **fk_kw)
                return sa.schema.Column(col_name, sa.types.Integer(), fk)

            gen_column.func_name = func_name
            return gen_column

        attr = get_column_attr(col_name, col_name, self.target_cls, target_col)
        yield col_name, sa.ext.declarative.declared_attr(attr)

        # relationship
        def get_rel_attr(func_name, target_cls, col_name):
            model_name = self.model

            def gen_relationship(cls):
                kw = dict(uselist=False, post_update=True)
                local = cls.__name__ + '.' + col_name
                remote = target_cls + '.id'
                if model_name == target_cls:
                    local = 'foreign({})'.format(local)
                    remote = 'remote({})'.format(remote)
                kw['primaryjoin'] = '{} == {}'.format(local, remote)

                if 'backref' in type_args:
                    backref_name = type_args['backref']
                    backref_kw = {}
                    if isinstance(backref_name, dict):
                        backref_kw.update(backref_name)
                        backref_name = backref_kw.pop('name')

                    kw['backref'] = sa.orm.backref(backref_name, **backref_kw)

                return sa.orm.relationship(target_cls, **kw)

            gen_relationship.func_name = func_name
            return gen_relationship

        attr = get_rel_attr(self.name, self.target_cls, col_name)
        yield self.name, sa.ext.declarative.declared_attr(attr)

    def _m2m_relationship(self, target_col, col_name, type_args):
        model_name = self.model

        def get_m2m_attr(func_name, target_cls, secondary_tbl_name=None):

            def gen_m2m_relationship(cls):
                is_self_referential = (model_name == target_cls)
                src_name = cls.__tablename__
                src_pk = src_name + '.id'
                src_col = src_name + '.c.id'
                target_name = func_name.lower()
                local_src_col = model_name.lower() + '_id'
                local_target_col = target_name + '_id'
                tbl_name = secondary_tbl_name

                if tbl_name is None:
                    tbl_name = (src_name + '_' + func_name)

                secondary_table = sa.Table(
                    tbl_name,
                    cls.metadata,
                    sa.Column(local_src_col, sa.ForeignKey(src_pk)),
                    sa.Column(local_target_col, sa.ForeignKey(target_col)),
                    sa.schema.UniqueConstraint(local_src_col,
                                               local_target_col),)

                rel_kw = dict(secondary=secondary_table)
                if is_self_referential:
                    tmpl = '{} == {}.c.{}'
                    rel_kw['primaryjoin'] = tmpl.format(src_col, tbl_name,
                                                        local_src_col)
                    rel_kw['secondaryjoin'] = tmpl.format(
                        target_cls + '.id', tbl_name, local_target_col)

                if 'backref' in type_args:
                    backref_name = type_args['backref']
                    backref_kw = {}
                    if isinstance(backref_name, dict):
                        backref_kw.update(backref_name)
                        backref_name = backref_kw.pop('name')

                    rel_kw['backref'] = sa.orm.backref(backref_name,
                                                       **backref_kw)

                return sa.orm.relationship(target_cls, **rel_kw)

            gen_m2m_relationship.func_name = func_name
            return gen_m2m_relationship

        attr = get_m2m_attr(self.name, self.target_cls)
        yield self.name, sa.ext.declarative.declared_attr(attr)


@form_field
class EntityFormField(FormField):
    ff_type = awbff.JsonSelect2Field

    def __init__(self, model, data, *args, **kwargs):
        super(EntityFormField, self).__init__(model, data, *args, **kwargs)
        self.module_endpoint = data.get('endpoint', self.name.lower())

    def get_type(self, *args, **kwargs):
        return (awbff.JsonSelect2MultipleField if self.multiple else
                awbff.JsonSelect2Field)

    def get_extra_args(self, *args, **kwargs):
        extra_args = super(EntityFormField, self).get_extra_args(*args,
                                                                 **kwargs)
        target = extra_args['model_class'] = self.data['target']
        ajax_endpoint = self.data.get('ajax_endpoint',
                                      target.lower() + '.json_search')
        extra_args['ajax_source'] = Endpoint(ajax_endpoint)
        extra_args['multiple'] = self.multiple
        return extra_args

    def setup_widgets(self, extra_args):
        self.setup_widgets_from_data(extra_args)

        if 'view_widget' not in extra_args:
            extra_args['view_widget'] = aw_widgets.EntityWidget()

        if 'widget' not in extra_args:
            extra_args['widget'] = aw_widgets.Select2Ajax(
                multiple=self.multiple)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from flask import current_app
from openpyxl.cell import STRING_TYPES

from .base import Column

__all__ = ('TagsColumn',)


class TagsColumn(Column):
    """
  Columns for :class:`abilian.core.models.tags.Tag` items
  """
    expected_cell_types = STRING_TYPES

    def data(self, item):
        ext = current_app.extensions.get('tags')
        if not ext or not ext.is_support_tagging(item):
            yield None, None

        value = sorted(ext.entity_tags(item))
        import_value = u'; '.join(unicode(t) for t in value) if value else u''
        yield import_value, value

    def deserialize(self, value):
        # type_ has already 'deserialized' value
        return value



# coding=utf-8
from __future__ import absolute_import

import setuptools
from distutils.command.build import build as _build
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.develop import develop as _develop

REQUIREMENTS = ['abilian-core', 'phonenumbers>=7.1.0', 'openpyxl>=2.3.3',]


class build(_build):
    sub_commands = [('compile_catalog', None)] + _build.sub_commands


class sdist(_sdist):
    sub_commands = [('compile_catalog', None)] + _sdist.sub_commands


class develop(_develop):

    def run(self):
        _develop.run(self)
        self.run_command('compile_catalog')


setuptools.setup(
    name='abilian-crm-core',
    use_scm_version=True,
    url='http://docs.abilian.com/',
    license='LGPL',
    author='Abilian SAS',
    author_email='contact@abilian.com',
    description='Core framework for CRM applications',
    packages=['abilian.crm'],
    zip_safe=False,
    platforms='any',
    setup_requires=['babel', 'setuptools-git', 'setuptools_scm'],
    install_requires=REQUIREMENTS,
    extras_require={},
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Flask',
    ],
    cmdclass={
        'build': build,
        'sdist': sdist,
        'develop': develop,
    })



# coding=utf-8
"""
"""
from __future__ import absolute_import

import datetime

from openpyxl.cell import NUMERIC_TYPES, STRING_TYPES, TIME_TYPES
from openpyxl.utils.datetime import from_excel

from .base import Column

__all__ = ('DateTimeColumn', 'DateColumn')


class DateTimeColumn(Column):
    """Column for datetime.datetime objects."""
    _date_fmt = '%Y-%m-%d %H:%M:%S'
    expected_cell_types = TIME_TYPES
    adapt_cell_types = TIME_TYPES + NUMERIC_TYPES
    _text_adapt = '%H:%M:%S %d/%m/%Y'

    # def serialize(self, value):
    #   return value.strftime(self._date_fmt)

    def _adapt_from_cell(self, value, workbook):
        if isinstance(value, STRING_TYPES):
            return datetime.datetime.strptime(value, self._text_adapt)

        if type(value) in NUMERIC_TYPES:
            max_year = datetime.datetime.now().year + 1
            if 2000 < value < max_year:
                # just current year
                return datetime.datetime(int(value), 1, 1)

            # default: cell is formatted as number in excel, but should be displayed
            # as date
            return from_excel(value)

        # just in case we reach here
        return value

    def deserialize(self, value):
        return datetime.datetime.strptime(value, self._date_fmt)


class DateColumn(DateTimeColumn):
    """Column for datetime.date objects."""
    _date_fmt = '%Y-%m-%d'
    _text_adapt = '%d/%m/%Y'

    def _adapt_from_cell(self, value, workbook):
        dt = DateTimeColumn._adapt_from_cell(self, value, workbook)
        return dt.date()

    def deserialize(self, value):
        return DateTimeColumn.deserialize(self, value).date()



# coding=utf-8
"""
"""
from __future__ import absolute_import

from openpyxl.cell import NUMERIC_TYPES, STRING_TYPES

from .base import Column

__all__ = ('EmptyColumn', 'TextIntegerColumn')


class EmptyColumn(Column):
    """
    Useful when an export needs a blank column.
    """
    importable = False

    def __init__(self, label=u''):
        super(EmptyColumn, self).__init__('', label=label, type_=None)

    def data(self, item):
        yield None, None


class TextIntegerColumn(Column):
    """
    Column for text that may be entered as integer, like a zipcode.
    """
    expected_cell_types = STRING_TYPES
    adapt_cell_types = NUMERIC_TYPES

    def _adapt_from_cell(self, value, cell_type, workbook):
        if isinstance(value, float):
            value = unicode(int(value))
        return value, value



# coding=utf-8
"""
"""
from __future__ import absolute_import

from .base import Column, ColumnSet, Invalid

from .dates import DateTimeColumn, DateColumn
from .vocabulary import VocabularyColumn
from .tags import TagsColumn
from .postaladdress import PostalAddressColumn
from .misc import EmptyColumn, TextIntegerColumn
from .related import RelatedColumnSet, ManyRelatedColumnSet
from .yearly import ManyYearlyColumnSet

__all__ = ("Column", "ColumnSet", "Invalid", "EmptyColumn", "DateTimeColumn",
           "DateColumn", "VocabularyColumn", "TagsColumn",
           "PostalAddressColumn", "TextIntegerColumn", "RelatedColumnSet",
           "ManyRelatedColumnSet", "ManyYearlyColumnSet")



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

from abilian.core.models import IdMixin, Model

from .jinja_filters import format_phonenumber


class PostalAddress(IdMixin, Model):
    __tablename__ = 'crm_postal_address'

    #: Multi-lines field
    street_lines = sa.Column(sa.UnicodeText)
    #: State / Province / Region / Military PO /...
    administrative_area = sa.Column(sa.UnicodeText)
    #: County / District
    sub_administrative_area = sa.Column(sa.UnicodeText)
    #: City / Town
    locality = sa.Column(sa.UnicodeText)
    #: Postal code / ZIP code
    postal_code = sa.Column(sa.UnicodeText)
    #: Country ISO code
    country = sa.Column(sa.UnicodeText, nullable=False)

    __table_args__ = (
        # we DO require a country
        sa.schema.CheckConstraint(sa.sql.func.length(country) > 0),)


class PhoneNumber(IdMixin, Model):
    __tablename__ = 'crm_phonenumbers'

    #: number type: mobile, pro... left as free text
    type = sa.Column(sa.UnicodeText,
                     default=u'',
                     server_default=sa.sql.text(u"''"))
    #: phone number
    number = sa.Column(sa.UnicodeText, nullable=False)

    def __unicode__(self):
        if not self.number:
            self.number = u''
        if not self.type:
            self.type = u''
        return u'%s: %s' % (self.type, format_phonenumber(self.number,
                                                          international=True))



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function

import pytest

from .base import assert_valid_identifier


def test_assert_valid_identifier():
    avi = assert_valid_identifier
    avi(u'a')
    avi(u'ab_C42')
    avi(u'_a1')

    invalid = (u'a test', u'été', u'a$' u'@var', u'newline\n',)

    for ident in invalid:
        with pytest.raises(ValueError):
            print(ident)
            avi(ident)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from abilian.crm.forms import PostalAddressForm
from abilian.crm.models import PostalAddress

from .related import RelatedColumnSet


class PostalAddressColumn(RelatedColumnSet):
    """
  Columns for :class:`abilian.crm.models.PostalAddress` items
  """

    def __init__(self, attr, label=None, type_=None, required=False):
        from .. import ExcelManager
        manager = ExcelManager(PostalAddress, PostalAddressForm)
        columns = manager.columns.columns
        super(PostalAddressColumn, self).__init__(attr, columns, label,
                                                  required)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

import abilian.web.forms.fields as awbff
import abilian.web.forms.widgets as aw_widgets

from .base import Field, FormField
from .registry import form_field, model_field


@model_field
class Vocabulary(Field):
    sa_type = sa.types.Integer
    default_ff_type = 'VocabularyFormField'

    def __init__(self, model, data, *args, **kwargs):
        super(Vocabulary, self).__init__(model, data, *args, **kwargs)
        self.voc_cls = data['vocabulary']['cls']

    def get_model_attributes(self, *args, **kwargs):
        relation_name = self.name
        attr_name = self.name + '_id'

        if not self.multiple:
            # column
            def get_column_attr(func_name, col_name, target_col):

                def gen_column(cls):
                    return sa.schema.Column(col_name,
                                            sa.ForeignKey(target_col,
                                                          ondelete='SET NULL'),)

                gen_column.func_name = func_name
                return gen_column

            attr = get_column_attr(attr_name, attr_name, self.voc_cls.id)
            yield attr_name, sa.ext.declarative.declared_attr(attr)

            # relationship
            def get_rel_attr(func_name, target_cls, attr_name):

                def gen_relationship(cls):
                    primary_join = '{} == {}'.format(
                        cls.__name__ + '.' + attr_name,
                        target_cls.__name__ + '.id')
                    return sa.orm.relationship(target_cls,
                                               primaryjoin=primary_join)

                gen_relationship.func_name = func_name
                return gen_relationship

            attr = get_rel_attr(relation_name, self.voc_cls, attr_name)
            yield relation_name, sa.ext.declarative.declared_attr(attr)

        else:  # m2m

            def get_m2m_attr(func_name, target_cls, secondary_tbl_name=None):

                def gen_m2m_relationship(cls):
                    src_name = cls.__tablename__
                    target_name = target_cls.__tablename__
                    tbl_name = secondary_tbl_name
                    if tbl_name is None:
                        tbl_name = (src_name + '_' + target_name)
                    src_col = cls.__name__.lower() + '_id'
                    secondary_table = sa.Table(
                        tbl_name,
                        cls.metadata,
                        sa.Column(src_col, sa.ForeignKey(src_name + '.id')),
                        sa.Column('voc_id', sa.ForeignKey(target_name + '.id')),
                        sa.schema.UniqueConstraint(src_col, 'voc_id'),)
                    return sa.orm.relationship(target_cls,
                                               secondary=secondary_table)

                gen_m2m_relationship.func_name = func_name
                return gen_m2m_relationship

            relation_secondary_tbl_name = \
                '{}_{}'.format(self.model.lower(), self.name.lower())

            rel_attr = get_m2m_attr(relation_name, self.voc_cls,
                                    relation_secondary_tbl_name)
            yield relation_name, sa.ext.declarative.declared_attr(rel_attr)


@form_field
class VocabularyFormField(FormField):
    ff_type = awbff.QuerySelect2Field

    def get_extra_args(self, *args, **kwargs):
        extra_args = super(VocabularyFormField, self) \
            .get_extra_args(*args, **kwargs)
        extra_args['multiple'] = self.multiple
        extra_args['get_label'] = 'label'

        def gen_voc_query(voc_cls):

            def query_voc():
                return voc_cls.query.active().all()

            query_voc.func_name = 'query_vocabulary_{}'.format(
                voc_cls.Meta.name)
            return query_voc

        voc_cls = self.data['vocabulary']['cls']
        extra_args['query_factory'] = gen_voc_query(voc_cls)
        return extra_args

    def setup_widgets(self, extra_args):
        extra_args['widget'] = aw_widgets.Select2(multiple=self.multiple,
                                                  unescape_html=True)
        extra_args['view_widget'] = aw_widgets.ListWidget()
